/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>

#include <algorithm>
#include <set.h>
#include <strstream.h>
#include <fstream.h>

#include <config.h>
#include <logf.h>
#include <stringtools.h>
#include <dboxversion.h>
#include <mtoolrfc.h>
#include <htmltools.h>
#include <nls.h>
#include "online.h"
#include "url.h"
#include "htmlview.h"
#include "writemail.h"
#include "basic_www.h"
#include "bookmarks.h"

bool do_basic_wwwt::go_back()
 {
  if(history_stack.empty())
   {
    end=TRUE;
    return TRUE;
   }

  current=history_stack.back();
  url=current.url;
  history_stack.pop_back();

  return FALSE;
 }

void do_basic_wwwt::mailto()
 {
  //writemail(host, "", "", (ifstream *)NULL);
 }

void do_basic_wwwt::historyt::clear()
 {
  page=0;
  page_has_link=FALSE;
  current_link=0;
 }

bool do_basic_wwwt::get_and_format_error()
 {
  string page, title, error;
  int error_number=0;
  if(path.c_str()[0]=='/') error_number=atoi(path.c_str()+1);

  if(host=="open.lookup")
   {
    title=nls["online.httperror.open.title"];
    error=nls["online.httperror.open.text"];
    error+=": ";
    switch(error_number)
     {
     case HOST_NOT_FOUND: error+=nls["online.gethostbyname.host_not_found"]; break;
     case NO_ADDRESS: error+=nls["online.gethostbyname.no_address"]; break;
     case NO_RECOVERY: error+=nls["online.gethostbyname.no_recovery"]; break;
     case TRY_AGAIN: error+=nls["online.gethostbyname.try_again"]; break;
     }
   }
  else if(host=="open.socket" || host=="read")
   {
    title=nls["online.httperror.open.title"];
    error=nls["online.httperror.open.text"];
    error+=": ";
    error+=strerror(error_number);
   }
  else if(host=="protocol")
   {
    title=nls["online.httperror.protocol.title"];
    error=nls["online.httperror.protocol.text"];
    if(path!="")
     {
      error+=": ";
      error+=htmlize_string(path.c_str()+1);
     }
   }
  else if(host=="Content-Type")
   {
    title=nls["online.httperror.content-type.title"];
    error=nls["online.httperror.content-type.text"];
    if(path!="")
     {
      error+=": ";
      error+=htmlize_string(path.c_str()+1);
     }
   }
  else if(host=="nntp")
   {
    title=nls["online.nntperror.title"];
    error=nls["online.nntperror.text"];
    if(path!="")
     {
      error+=": ";
      error+=htmlize_string(path.c_str()+1);
     }
   }
  else if(host=="ftp")
   {
    title=nls["online.ftperror.title"];
    error=nls["online.ftperror.text"];
    if(path!="")
     {
      error+=": ";
      error+=htmlize_string(path.c_str()+1);
     }
   }
  else
   {
    title=host;
    error="unknown";
   }

  source="<html><head><title>";
  source+=nls["online.error"];
  source+=": ";
  source+=title;
  source+="</title></head>\n";
  source+="<body>\n";
  source+=nls["online.httperror.text"];
  source+="<p>\n<blockquote>\n";
  source+=error;
  source+="</blockquote>\n</body></html>";
  
  return format_html(source);
 }

bool do_basic_wwwt::get_and_format_http()
 {
  ibsockett socket;
  word port_nr;

  port_nr=atol(port.c_str());
  if(port_nr==0) port_nr=80;

  if(socket.open("HTTP", host, port_nr))
   {
    string new_url="error://";

    if(socket.get_error_type()==sockett::SOCKET_ERROR)
      new_url+="open.socket/";
    else
      new_url+="open.lookup/";

    new_url+=i2string(socket.get_error_number())+"/"+url;
    parse_url(new_url, protocol, host, port, path);
    return get_and_format_error();
   }

  string http_request;

  http_request="GET ";
  http_request+=path;
  http_request+=" HTTP/1.0\r\n";
  http_request+="Connection: close\r\n";
  http_request+="Host: "+dn_string(host);
  if(port_nr!=80) http_request+=":"+port;
  http_request+="\r\n";
  http_request+="User-Agent: DEBRIS/" DBOXVERSIONNR "\r\n";
  http_request+="\r\n";

  socket.write(http_request.c_str(), http_request.size());

  string page, status_response;

  char line[1000];
  bool first_line=TRUE;

  do
   {
    if(socket.gets(line, 999)==NULL)
     {
      string new_url="error://read/"+
                     i2string(socket.get_error_number())+"/"+url;
      parse_url(new_url, protocol, host, port, path);
      return get_and_format_error();
     }

    if(first_line)
     {
      status_response=line;
      first_line=FALSE;
     }
    else
      page+=line;
   }
  while(line[0]!=0 && line[0]!='\r' && line[0]!='\n');

  rfcmessaget msg;
  msg.data=page.c_str();
  msg.endofmail=msg.endoffile=page.c_str()+page.size();
  msg.parse();

  string content_length;
  dword length=0;
  if(!msg.findsingleheader(content_length, "Content-Length"))
    length=atol(content_length.c_str());

   {
    char *buffer=(char *)malloc(MAX_SIZE);
    if(buffer==(char *)NULL) 
      logf.internalerror("get_and_format_http", "out of memory");
    if(length!=0) socket.set_progress_bar(TRUE, length);
    size_t size=socket.read(buffer, MAX_SIZE);
    page.append(buffer, size);
    free(buffer);
   }

  msg.data=page.c_str();
  msg.endofmail=msg.endoffile=page.c_str()+page.size();
  msg.parse();

  return format_http_msg(status_response, msg);
 }

bool do_basic_wwwt::format_http(const string &http)
 {
  unsigned i;
  string status_response;

  for(i=0; i<http.size() && http[i]!='\n'; i++);
  if(i==http.size()) 
    return TRUE;

  status_response.assign(http, 0, i);

  rfcmessaget msg;
  msg.data=http.c_str()+i+1;
  msg.endofmail=msg.endoffile=http.c_str()+http.size();
  msg.parse();
  return format_http_msg(status_response, msg);
 }

bool do_basic_wwwt::format_http_msg(const string &status_response,
                                    const rfcmessaget &msg)
 {
  unsigned code, i;

  for(i=0; i<status_response.size() && status_response[i]!=' '; i++);

  if(i<status_response.size()) 
    code=atoi(status_response.c_str()+i);
  else
    code=200;

  string location;
  if((code==301 || code==302) &&
     !msg.findsingleheader(location, "Location") && 
     redirection<5)
   {
    string::size_type pos=location.find_first_of(":/");
    if(pos==string::npos || location[pos]!=':') // relative
      make_url(url, protocol, host, port, location);
    else
      url=location;
    parse_url(url, protocol, host, port, path);
    redirection++;
    return get_and_format();
   }

  string content_type;
  if(!msg.findsingleheader(content_type, "Content-Type"))
   {
    mimeheadert mime_header;
    parse_mime_header(content_type, mime_header);

    if(stringcasecmp(mime_header.value, "text")==0 &&
       stringcasecmp(mime_header.subvalue, "html")==0)
     {
     }
    else if((stringcasecmp(mime_header.value, "text")==0 &&
             stringcasecmp(mime_header.subvalue, "plain")==0) ||
             stringcasecmp(mime_header.value, "message")==0)
     {
      source.assign(msg.data+msg.headerlen, msg.len);
      string filtered_source="<html><body><pre>"+
        htmlize_string(source)+"</pre></body></html>";

      return format_html(filtered_source);
     }
    else
      return content_type_not_builtin(mime_header, msg);
   }

  source.assign(msg.data+msg.headerlen, msg.len);

  return format_html(source);
 }

bool do_basic_wwwt::content_type_not_builtin
  (mimeheadert content_type, const rfcmessaget &msg)
 {
  string new_url="error://Content-Type/"+url;
  parse_url(new_url, protocol, host, port, path);
  return get_and_format_error();
 }

bool do_basic_wwwt::get_and_format_source()
 {
  string page, original_url;
  if(path!="") original_url=path.c_str()+1;

  page="<html><head><title>Source of "+htmlize_string(original_url)+
       "</title></head>\n"
       "<body><pre>"+htmlize_string(source)+"</pre>\n</body>\n";

  return format_html(page);
 }

bool do_basic_wwwt::format_html(const string &page)
 {
  istrstream in(page.c_str(), page.size());
  htmlformat=new htmlformatt;
  htmlformat->set_width(online.sizex);
  htmlformat->format(in);
  return FALSE;
 }

bool do_basic_wwwt::get_and_format_internal()
 {
  if(host=="history")
   {
    string page="<html><head><title>History</title></head>\n"
      "<body><center><h1>History</h1></center><hr><table>\n";

    for(unsigned i=0; i<history_stack.size(); i++)
     {
      page+="<tr><th>"+i2string(i+1)+".</th>\n<td><a href=\""+
            history_stack[i].url+"\">";
         
      if(history_stack[i].title=="")
        page+="no title";
      else
        page+=htmlize_string(history_stack[i].title);

      page+="</a><br>"+htmlize_string(history_stack[i].url);
      page+="</td>";
     }

    page+="</table>\n</body></html>\n";

    return format_html(page);
   }
  else if(host=="bookmarks")
   {
    string page;
    get_bookmarks(page);
    return format_html(page);
   }
  
  return TRUE;
 }

bool do_basic_wwwt::get_and_format2()
 {
  if(protocol=="http")
    return get_and_format_http();
  else if(protocol=="ftp")
    return get_and_format_ftp();
  else if(protocol=="news")
    return get_and_format_nntp();
  else if(protocol=="internal")
    return get_and_format_internal();
  else if(protocol=="error")
    return get_and_format_error();
  else if(protocol=="source")
    return get_and_format_source();

  url="error://protocol/"+protocol;
  parse_url(url, protocol, host, port, path);
  return get_and_format_error();
 }

bool do_basic_wwwt::get_and_format()
 {
  bool protocol_cacheable=
    (protocol=="http" ||
     protocol=="ftp" ||
     protocol=="news");

  if(protocol_cacheable)
   {
    map<string,cache_entryt>::iterator it;
    it=cache.find(url);
    if(it!=cache.end())
     {
      htmlformat=new htmlformatt(it->second.htmlformat);
      return FALSE;
     }
   }

  if(get_and_format2())
    return TRUE;

  if(protocol_cacheable && source.size()<200000)
   {
    cache_entryt entry;
    entry.htmlformat=*htmlformat;
    cache.insert(pair<string,cache_entryt>(url, entry));
   }

  return FALSE;
 }

void do_basic_wwwt::do_www(const string &_url)
 {
  url=_url;

  for(end=FALSE; !end;)
   {
    if(url=="") url=config.defaulturl;
    parse_url(url, protocol, host, port, path);
    dn_string(protocol);
    redirection=0;

    if(protocol=="mailto")
     {
      mailto();
      go_back();
     }
    else if(get_and_format())
      go_back();
    else
     {
      string new_url=url;
      htmlviewt htmlview;
      htmlview.page=current.page;
      htmlview.page_has_link=current.page_has_link;
      htmlview.current_link=current.current_link;
      if(htmlview.view(*htmlformat, new_url))
        end=TRUE;
      else
       {
        if(new_url=="BACK")
          go_back();
        else if(new_url=="ADD_BOOKMARK")
         {
          // TODO
         }
        else
         {
          if(protocol!="internal" || host!="history")
           {
            historyt history;
            history.url=url;
            history.title=htmlformat->title;
            history.page=htmlview.page;
            history.current_link=htmlview.current_link;
            history.page_has_link=htmlview.page_has_link;
            history_stack.push_back(history);
           }

          current.clear();

          if(new_url=="HELP")
            url="internal://help/internet/www/";
          else if(new_url=="HISTORY")
            url="internal://history/";
          else if(new_url=="BOOKMARKS")
            url="internal://bookmarks/";
          else if(new_url=="SOURCE")
            url="source://source/"+url;
          else if(new_url=="HOME")
            url=config.defaulturl;
          else
            url=new_url;
         }
       }

      delete htmlformat;
     }
   }
 }

do_basic_wwwt::do_basic_wwwt()
 {
 }

do_basic_wwwt::~do_basic_wwwt()
 {
 }
