/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <dirent.h>
#include <fcntl.h>
#include <time.h>
#include <sys/stat.h>

#include <minmax.h>

#include <nls.h>
#include <ui.h>
#include "online.h"
#include "editor.h"
#include "help.h"

#define MAXLINES 1000
#define MAXWIDTH 80

class doeditt
 {
 public:
  dword x, y, scrolly, anzlines;

  time_t cposdraw_last;
  bool cposdraw_skipped;
  byte cposdraw_count;

  const char *title;
  const char *filename;

  void save();
  void load();
  bool doedit(word startline);
  void drawtitle();
  void drawscreen();
  void drawline(dword line);
  void drawcpos(bool forcedraw=FALSE);
  void setcursor();
  byte really_quit(); // 0 = Abbruch 1 = speichern 2 = nicht speichern
  void dohelp();

  doeditt()
   {
    title=filename=NULL;
    x=y=scrolly=anzlines=0;
    cposdraw_last=0;
    cposdraw_skipped=FALSE;
    cposdraw_count=0;
   }

  struct linet
   {
    char data[MAXWIDTH];
   } lines[MAXLINES];

 private:
  void dodown();
  void doup();
  bool dokeyent();
  void dokeyend();
  void dodelline();
 };

void doeditt::dohelp()
 {
  displayhelp("editor");
  drawscreen();
 }

void doeditt::drawcpos(bool forcedraw)
 {
  time_t t;

  time(&t);

  if(!forcedraw && cposdraw_last==t)
   {
    if(cposdraw_count>3)
     {
      cposdraw_skipped=TRUE;
      return;
     }
    else
      cposdraw_count++;
   }
  else
    cposdraw_count=0;

  cposdraw_last=t;
  cposdraw_skipped=FALSE;

  online.setcursor(71, 1);
  online.setattr(10, FALSE);
  online.printf("%4lu %4lu ", x+1, y+1);
  online.setattr(1, FALSE);
 }

void doeditt::drawtitle()
 {
  byte xpos;

  online.setcursor(1, 1);
  online.setattr(10, FALSE);

  unsigned len=strlen(title);
  if(len>(online.sizex-30))
   {
    len=online.sizex-30;
    online.printf(" %1.-*s", len, title);
   }
  else
    online.printf(" %s", title);

  xpos=1+len;

  for(; xpos<=online.sizex-19; xpos++) online.putch(' ');
 
  online.printf("F1: Hilfe", x+1, y+1);

  drawcpos(TRUE);

  online.setattr(1, FALSE);
 }

byte doeditt::really_quit()
 {
  online.setattr(2, FALSE);

  online.box(online.sizex/2-20, 10, online.sizex/2+20, 16,
             BOXFRAMEDBL, NULL, TRUE);

   {
    const char *msg=nls["online.savetext"];
    online.setcursor((online.sizex-strlen(msg))/2, 12);
    online.printf("%s", msg);
   }

  menut menu(&online);
  menu.x=online.sizex/2-20;
  menu.y=10;

  buttonmet yesme;
  yesme.x=6;
  yesme.y=4;
  yesme.id=0;
  yesme.selectkey=Key_C_Ent;
  yesme.sizex=6;
  yesme.text=nls["online.yes"];
  menu.add(&yesme);

  buttonmet nome;
  nome.x=23;
  nome.y=4;
  nome.id=1;
  nome.sizex=6;
  nome.text=nls["online.no"];
  menu.add(&nome);

  switch(menu.menu())
   {
   case 0:
    return 1;

   case 1:
    return 2;

   case MENU_HELP:
    dohelp();
   }

  return 0;
 }

void doeditt::setcursor()
 {
  online.setcursor(x+1, y+2-scrolly);
 }

void doeditt::save()
 {
  dword line;
  FILE *outfile;

  outfile=fopen(filename, "wb");

  for(line=0; line<anzlines; line++)
    fprintf(outfile, "%s\r\n", lines[line].data);

  fclose(outfile);
 }

void doeditt::load()
 {
  FILE *infile;
  char tempstr[TEXTLEN], *tptr;
  dword line;

  for(y=0; y<MAXLINES; y++)
    lines[y].data[0]=0;

  anzlines=x=y=scrolly=0;

  infile=fopen(filename, "rt");
  if(infile==NULL) return;

  line=0;

  while(line<MAXLINES && fgets(tempstr, TEXTLEN-1, infile)!=NULL)
   {
    tptr=strrchr(tempstr, '\r'); if(tptr!=NULL) *tptr=0;
    tptr=strrchr(tempstr, '\n'); if(tptr!=NULL) *tptr=0;

    strmaxcpy(lines[line].data, tempstr, MAXWIDTH-1);
    line++;
   }

  anzlines=line;

  fclose(infile);
 }

void doeditt::drawline(dword line)
 {
  online.setcursor(1, line+2-scrolly);
  online.cleareol();
  if(line>=MAXLINES) return;
  if(lines[line].data==NULL) return;

  if(line==anzlines)
   {
    online.setattr(1, TRUE);
    online.printf("EOF");
    online.setattr(1, FALSE);
   }
  else
   {
    char tempstr[TEXTLEN*2];

    strmaxcpy(tempstr, lines[line].data, TEXTLEN-1);
    tempstr[online.sizex-1]=0;
    online.printf("%s", tempstr);
   }
 }

void doeditt::drawscreen()
 {
  byte line;

  online.setattr(1, FALSE);
  online.clrscr();
  drawtitle();

  for(line=0; line<online.sizey-2; line++)
    drawline(line+scrolly);

  drawstatusbar();

  setcursor();
 }

void doeditt::doup()
 {
  if(y!=0)
   {
    y--;
    if(y<scrolly)
     {
      scrolly--;
      online.setcursor(1, online.sizey-1);
      online.deleteline();
      online.setcursor(1, 2);
      online.insertline();
      drawline(y);
     }

    drawcpos();
    drawstatusbar();
    setcursor();
   }
 }

void doeditt::dodown()
 {
  if(y<MAXLINES-2 && y<anzlines)
   {
    y++;
    if(y>(scrolly+online.sizey-3))
     {
      if(status.online2)
       {
        online.console.setcursor(1, online.console.sizey);
        online.console.deleteline();
       }

      scrolly++;
      online.setcursor(1, 2);
      online.deleteline();
      drawline(y);
     }
    else if(y==anzlines)
      drawline(y);

    drawcpos();
    drawstatusbar();
    setcursor();
   }
 }

bool doeditt::dokeyent()
 {
  if(y>=MAXLINES-2)
    return TRUE;

  if(y==anzlines)
   {
    online.setcursor(1, y+2-scrolly);
    online.cleareol();
    anzlines++;
    dodown();
   }
  else if(y==anzlines-1)
   {
    if(x<strlen(lines[y].data))
     {
      strcpy(lines[y+1].data, lines[y].data+x);
      lines[y].data[x]=0;
      online.cleareol();
      x=0;
      anzlines++;
      dodown();
      drawline(y);
     }
    else
     {
      x=0;
      anzlines++;
      dodown();
      online.setcursor(1, y+2-scrolly);
      online.cleareol();
     }

    if(y<(scrolly+online.sizey-3))
      drawline(y+1);

    drawcpos();
    setcursor();
   }
  else if(y<MAXLINES-2)
   {
    unsigned i;

    for(i=anzlines; i>(y+1); i--)
      lines[i]=lines[i-1];

    anzlines++;

    if(x<strlen(lines[y].data))
     {
      strcpy(lines[y+1].data, lines[y].data+x);
      lines[y].data[x]=0;
     }
    else
      lines[y+1].data[0]=0;

    x=0;
    dodown();

    for(i=y-scrolly-1; i<unsigned(online.sizey-2); i++)
      drawline(i+scrolly);

    drawcpos();
    setcursor();
   }

  return FALSE;
 }

void doeditt::dokeyend()
 {
  if(x!=min(strlen(lines[y].data), unsigned(online.sizex-2)))
   {
    x=min(strlen(lines[y].data), unsigned(online.sizex-2));
    drawcpos();
    setcursor();
   }
 }

void doeditt::dodelline()
 {
  if(y==anzlines || anzlines==0) return;

  if(y==(anzlines-1))
   {
    lines[y].data[0]=0;
    anzlines--;

    drawline(y);

   if(y<(online.sizey+scrolly-3))
      drawline(y+1);

    setcursor();
   }
  else
   {
    unsigned i;

    anzlines--;

    for(i=y; i<anzlines; i++)
      strcpy(lines[i].data, lines[i+1].data);

    lines[anzlines].data[0]=0;

    for(i=y-scrolly; i<unsigned(online.sizey-2); i++)
      drawline(i+scrolly);

    setcursor();
   }
 }

bool doeditt::doedit(word startline)
 {
  word key;

  load();

  y=startline;
  if(y>anzlines) y=anzlines;
  if(y>(scrolly+online.sizey-3))
    scrolly=y-online.sizey;

 startover:
  drawscreen();

 schleife:
  switch(key=online.doreadkey())
   {
   case 0:
    if(cposdraw_skipped) 
     {
      drawcpos(TRUE);
      setcursor();
     }
    break;

   case Key_F1:
    dohelp();
    break;

   case Key_C_Z:
   case Key_Esc:
    switch(really_quit())
     {
     case 0:
      break;

     case 1:
      save();
      return FALSE;

     case 2:
      return TRUE;
     }
    if(status.dologout) return TRUE;
    goto startover;

   case Key_Up:
    doup();
    goto schleife;

   case Key_Dwn:
    dodown();
    goto schleife;

   case Key_Rgt:
    if(x<unsigned(online.sizex-2) && x<unsigned(MAXWIDTH-1))
     {
      x++;
      drawcpos();
      setcursor();
     }
    goto schleife;

   case Key_Lft:
    if(x!=0)
     {
      x--;
      drawcpos();
      setcursor();
     }
    goto schleife;

   case Key_End:
   case Key_C_E:
    dokeyend();
    goto schleife;

   case Key_Home:
   case Key_C_A:
    if(x!=0)
     {
      x=0;
      drawcpos();
      setcursor();
     }
    goto schleife;

   case Key_C_R:
    drawscreen();
    goto schleife;

   case Key_C_Y:
    dodelline();
    goto schleife;

   case Key_PgUp:
    if(y<=unsigned(online.sizey-3))
     {
      y=0;
      if(scrolly==0)
       {
        drawcpos();
        setcursor();
       }
      else
       {
        scrolly=0;
        drawscreen();
       }
     }
    else
     {
      y-=online.sizey-3;
      if(scrolly<unsigned(online.sizey-3))
        scrolly=0;
      else
        scrolly-=online.sizey-3;
      drawscreen();
     }
    goto schleife;

   case Key_PgDn:
    if(y+online.sizey-3>=anzlines)
     {
      y=anzlines;
      if(y>(scrolly+online.sizey-3))
       {
        scrolly+=online.sizey-3;
        drawscreen();
       }
      else
       {
        drawcpos();
        setcursor();
       }
     }
    else
     {
      y+=online.sizey-3;
      scrolly+=online.sizey-3;
      drawscreen();
     }
    goto schleife;

   case Key_Ent:
    dokeyent();
    goto schleife;

   case Key_Del:
   case Key_BS:
    if(y==anzlines)
     {
      if(y!=0)
       {
        doup();
        dokeyend();
        goto schleife;
       }
     }
    else if(x==0)
     {
     }
    else
     {
      if(x<=strlen(lines[y].data))
       {
        online.putch(8);

        for(unsigned i=x-1; i<strlen(lines[y].data); i++)
         {
          lines[y].data[i]=lines[y].data[i+1];
          online.printf("%c", lines[y].data[i]);
         }

        online.putch(' ');
       }

      x--;
      drawcpos();
      setcursor();
     }
    goto schleife;

   case 0xffff:
    return TRUE;

   default:
    if(lo(key)>=' ')
     {
      if(y==anzlines)
       {
        if(y==MAXLINES-2) goto schleife;

        anzlines++;

        if(y<(online.sizey+scrolly-3))
          drawline(y+1);

        online.setcursor(1, y+2-scrolly);
        online.cleareol();
        setcursor();
       }

      if(!(x<unsigned(online.sizex-2) && x<(MAXWIDTH-1)))
       {
        char *tptr;

        if(lo(key)==' ')
         {
          dokeyent();
          goto schleife;
         }

        tptr=strrchr(lines[y].data, ' ');
        if(tptr==NULL) goto schleife;
        x=tptr-lines[y].data+1;
        setcursor();
        if(dokeyent()) goto schleife;
        dokeyend();
       }

      if(x<unsigned(online.sizex-2) && x<(MAXWIDTH-1))
       {
        char line[MAXWIDTH*2];

        strmaxcpy(line, lines[y].data, (MAXWIDTH*2)-1);
        if(x==strlen(line))
         {
          line[x]=lo(key);
          line[x+1]=0;
          online.printf("%c", lo(key));
         }
        else if(x<strlen(line))
         {
          unsigned i;
          for(i=x; i<=strlen(lines[y].data); i++)
            line[i+1]=lines[y].data[i];

          line[x]=lo(key);

          for(i=x; i<=unsigned(online.sizex-2) && line[i]!=0; i++)
            online.printf("%c", line[i]);
         }
        else
         {
          for(unsigned i=strlen(line); i<x; i++)
            line[i]=' ';

          line[x]=lo(key);
          line[x+1]=0;
          online.printf("%c", lo(key));
         }

        strmaxcpy(lines[y].data, line, MAXWIDTH-1);
        x++;
        drawcpos();
        setcursor();
       }
     }
    goto schleife;
   }
  goto schleife;
 }

bool doedit(const char *filename, const char *title, word startline)
 {
  doeditt de;

  de.filename=filename;
  de.title=title;

  return de.doedit(startline);
 }
