/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <ctype.h>

#include <libdbox.h>
#include <charsettools.h>
#include <stringtools.h>

#include "htmlformat.h"

bool search(const vector<tag_paramt> &params, const string &id, 
            string &value)
 {
  value="";

  for(unsigned i=0; i<params.size(); i++)
    if(stringcasecmp(id, params[i].id)==0)
     {
      value=params[i].value;
      return FALSE;
     }
  
  return TRUE;
 }

const string &search(const vector<tag_paramt> &params, const string &id)
 {
  for(unsigned i=0; i<params.size(); i++)
    if(stringcasecmp(id, params[i].id)==0)
      return params[i].value;

  static string empty_string;  
  return empty_string;
 }

void htmlformatt::html_comment(istream &in)
 {
  char ch;
  int state=0;

  while(in.read(&ch, 1))
    if(state==0 || state==1)
     {
      if(ch=='-')
        state++;
      else
        state=0;
     }
    else // state==2
     {
      if(ch=='>')
        return;
      else if(ch!='-')
        state=0;
     }
 }

void htmlformatt::tag_params(istream &in, vector<tag_paramt> &params)
 {
  char ch;
  bool nextparam=TRUE, value=FALSE, quoted=FALSE;
  tag_paramt empty_param;

  while(in.read(&ch, 1))
   {
    if(quoted)
     {
      if(ch=='"')
        quoted=FALSE;
      else
       {
        if(params.empty())
          params.push_back(empty_param);

        params.back().value+=ch;
       }
     }
    else
      switch(ch)
       {
       case '>':
        return;

       case '"':
        quoted=TRUE;
        break;

       case '=':
        value=TRUE;
        break;

       case '\r':
       case '\t':
       case '\n':
       case ' ':
        nextparam=TRUE;
        break;

       default:
        if(nextparam || params.empty())
         {
          params.push_back(empty_param);
          value=nextparam=FALSE;
         }

        if(value)
          params.back().value+=ch;
        else
          params.back().id+=ch;
       }
   }
 }

void htmlchart::clear()
 {
  ibm_charset=islink=bold=underline=italic=preformat=islink=FALSE;
  center=marked=FALSE;
  ch=0;
  link=0;
 }

void htmlformatt::do_tag(istream &in,
                         const string &name, 
                         const vector <tag_paramt> &params)
 {
  if(stringcasecmp(name, "h1")==0 ||
     stringcasecmp(name, "h2")==0 ||
     stringcasecmp(name, "h3")==0 ||
     stringcasecmp(name, "h4")==0 ||
     stringcasecmp(name, "h5")==0 ||
     stringcasecmp(name, "h6")==0)
   {
    line_break();
   }
  else if(stringcasecmp(name, "/h1")==0 ||
     stringcasecmp(name, "/h2")==0 ||
     stringcasecmp(name, "/h3")==0 ||
     stringcasecmp(name, "/h4")==0 ||
     stringcasecmp(name, "/h5")==0 ||
     stringcasecmp(name, "/h6")==0)
   {
    line_break();
   }
  else if(stringcasecmp(name, "frame")==0)
   {
    line_break();
    linkt link;
    link.href=search(params, "src");
    links.push_back(link);
    attr_link=(link.href!="");
    text("Frame");
    string name=search(params, "name");
    if(name!="") text(": "+name);
    line_break();
    attr_link=FALSE;
   }
  else if(stringcasecmp(name, "img")==0)
    entity_text(search(params, "alt"));
  else if(stringcasecmp(name, "script")==0)
   {
    flush_word();
    attr_script=TRUE;
   }
  else if(stringcasecmp(name, "/script")==0)
   {
    flush_word();
    attr_script=FALSE;
   }
  else if(stringcasecmp(name, "/head")==0)
   {
    flush_word();
    attr_script=FALSE;
   }
  else if(stringcasecmp(name, "head")==0)
   {
    flush_word();
   }
  else if(stringcasecmp(name, "ul")==0)
   {
    line_break();
    indent_stack.push_back(-1);
   }
  else if(stringcasecmp(name, "blockquote")==0)
   {
    line_break();
    indent_stack.push_back(-2);
    indent=indent_stack.size()*2;
   }
  else if(stringcasecmp(name, "ol")==0)
   {
    line_break();
    indent_stack.push_back(1);
   }
  else if(stringcasecmp(name, "li")==0)
   {
    line_break();
    if(indent_stack.size()!=0)
     {
      indent=(indent_stack.size()-1)*2;
      if(indent_stack.back()==-1)
        entity_text("* ");
      else
       {
        char tempstr[TEXTLEN];
        snprintf(tempstr, TEXTLEN, "%u ", indent_stack.back());
        text(tempstr);
        indent_stack.back()++;
       }
      indent=indent_stack.size()*2;
     }
   }
  else if(stringcasecmp(name, "/ul")==0)
   {
    if(!indent_stack.empty())
     {
      indent_stack.pop_back();
      indent=indent_stack.size()*2;
     }
    line_break();
   }
  else if(stringcasecmp(name, "/blockquote")==0)
   {
    line_break();
    if(!indent_stack.empty())
     {
      indent_stack.pop_back();
      indent=indent_stack.size()*2;
     }
   }
  else if(stringcasecmp(name, "/ol")==0)
   {
    if(!indent_stack.empty())
     {
      indent_stack.pop_back();
      indent=indent_stack.size()*2;
     }
    line_break();
   }
  else if(stringcasecmp(name, "b")==0)
    attr_bold=TRUE;
  else if(stringcasecmp(name, "/b")==0)
    attr_bold=FALSE;
  else if(stringcasecmp(name, "i")==0)
    attr_italic=TRUE;
  else if(stringcasecmp(name, "/i")==0)
    attr_italic=FALSE;
  else if(stringcasecmp(name, "u")==0)
    attr_underline=TRUE;
  else if(stringcasecmp(name, "/u")==0)
    attr_underline=FALSE;
  else if(stringcasecmp(name, "a")==0)
   {
    linkt link;
    link.href=search(params, "href");
    attr_link=(link.href!="");
    links.push_back(link);
   }
  else if(stringcasecmp(name, "/a")==0)
   {
    attr_link=FALSE;
   }
  else if(stringcasecmp(name, "title")==0)
   {
    flush_word();
    attr_title=TRUE;
   }
  else if(stringcasecmp(name, "/title")==0)
   {
    flush_word();
    attr_title=FALSE;
   }
  else if(stringcasecmp(name, "form")==0)
   {
    flush_word();
    form(params);
   }
  else if(stringcasecmp(name, "input")==0)
   {
    flush_word();
    form_input(params);
   }
  else if(stringcasecmp(name, "/form")==0)
   {
    flush_word();
   }
  else if(stringcasecmp(name, "center")==0)
   {
    line_break();
    attr_center=TRUE;
   }
  else if(stringcasecmp(name, "/center")==0)
   {
    line_break();
    attr_center=FALSE;
   }
  else if(stringcasecmp(name, "table")==0)
   {
    line_break();
    attr_table=TRUE;
    table.clear();
    table.set_params(params);
   }
  else if(stringcasecmp(name, "/table")==0)
    render_table();
  else if(stringcasecmp(name, "tr")==0)
    do_tr(params);
  else if(stringcasecmp(name, "/tr")==0)
    line_break();
  else if(stringcasecmp(name, "td")==0)
    do_td(in, params, FALSE);
  else if(stringcasecmp(name, "th")==0)
    do_td(in, params, TRUE);
  else if(stringcasecmp(name, "/td")==0)
    flush_word();
  else if(stringcasecmp(name, "/th")==0)
    flush_word();
  else if(stringcasecmp(name, "p")==0)
   {
    if(lines.size()!=0 && lines[0].size()!=0)
     {
      line_break();
      new_line();
      //new_line();
     }

    string align=search(params, "align");
    if(align=="right")
      current_alignment=RIGHT;
    else if(align=="left")
      current_alignment=LEFT;
    else if(align=="center")
      current_alignment=CENTER;
    else
      current_alignment=default_alignment;
   }
  else if(stringcasecmp(name, "div")==0)
   {
    line_break();

    string align=search(params, "align");
    if(align=="right")
      current_alignment=RIGHT;
    else if(align=="left")
      current_alignment=LEFT;
    else if(align=="center")
      current_alignment=CENTER;
    else
      current_alignment=default_alignment;
   }
  else if(stringcasecmp(name, "/p")==0)
   {
    line_break();
    current_alignment=default_alignment;
   }
  else if(stringcasecmp(name, "/div")==0)
   {
    line_break();
    current_alignment=default_alignment;
   }
  else if(stringcasecmp(name, "br")==0)
   {
    flush_word();
    new_line();    
    //line_break();
   }
  else if(stringcasecmp(name, "hr")==0)
   {
    flush_word();
    if(lines.empty() || !lines[line].empty()) new_line();    

    if(!do_get_max_width)
     {
      htmlchart htmlchar(196);
      htmlchar.ibm_charset=TRUE;

      for(unsigned i=0; i<width; i++)
        lines[line].push_back(htmlchar);

      new_line();
     }
   }
  else if(stringcasecmp(name, "pre")==0)
   {
    flush_word();
    attr_preformat=TRUE;
   }
  else if(stringcasecmp(name, "/pre")==0)
   {
    flush_word();
    new_line();
    attr_preformat=FALSE;
   }
  else if(stringcasecmp(name, "DBOX-CHARSET-0")==0)
    attr_ibm_charset=TRUE;
  else if(stringcasecmp(name, "DBOX-CHARSET-1")==0)
    attr_ibm_charset=FALSE;
 }

void htmlformatt::get_tag(istream &in, string &tag_name, 
                          vector<tag_paramt> &params)
 {
  char ch;

  params.clear();
  tag_name="";

  while(in.read(&ch, 1))
   {
    switch(ch)
     {
     case '>':
      return;

     case '\r':
     case '\n':
     case '\t':
     case ' ':
      tag_params(in, params);
      return;

     default:
      tag_name+=ch;
      if(tag_name=="!--") 
       {
        html_comment(in);
        return;
       }
     }
   }
 }

void htmlformatt::tag(istream &in)
 {
  string tag_name;
  vector<tag_paramt> params;

  get_tag(in, tag_name, params);
  do_tag(in, tag_name, params);
 }

struct entityt
 {
  char *name;
  char code;
 } entity_table[]=
 {
   { "quot", '"' },
   { "AElig", 198 },
   { "Aacute", 193 },
   { "Acirc", 194 },
   { "Agrave", 192 },
   { "Aring", 197 },
   { "Atilde", 195 },
   { "Auml", 196 },
   { "Ccedil", 199 },
   { "ETH", 208 },
   { "Eacute", 201 },
   { "Ecirc", 202 },
   { "Egrave", 200 },
   { "Euml", 203 },
   { "Iacute", 205 },
   { "Icirc", 206 },
   { "Igrave", 204 },
   { "Iuml", 207 },
   { "Ntilde", 209 },
   { "Oacute", 211 },
   { "Ocirc", 212 },
   { "Ograve", 210 },
   { "Oslash", 216 },
   { "Otilde", 213 },
   { "Ouml", 214 },
   { "THORN", 222 },
   { "Uacute", 218 },
   { "Ucirc", 219 },
   { "Ugrave", 217 },
   { "Uuml", 220 },
   { "Yacute", 221 },
   { "aacute", 225 },
   { "acirc", 226 },
   { "aelig", 230 },
   { "agrave", 224 },
   { "amp", '&' },
   { "aring", 229 },
   { "atilde", 227 },
   { "auml", 228 },
   { "ccedil", 231 },
   { "eacute", 233 },
   { "ecirc", 234 },
   { "egrave", 232 },
   { "eth", 240 },
   { "euml", 235 },
   { "gt", 62 },
   { "iacute", 237 },
   { "icirc", 238 },
   { "igrave", 236 },
   { "iuml", 239 },
   { "lt", '<' },
   { "ntilde", 241 },
   { "oacute", 243 },
   { "ocirc", 244 },
   { "ograve", 242 },
   { "oslash", 248 },
   { "otilde", 245 },
   { "ouml", 246 },
   { "szlig", 223 },
   { "thorn", 254 },
   { "uacute", 250 },
   { "ucirc", 251 },
   { "ugrave", 249 },
   { "uuml", 252 },
   { "yacute", 253 },
   { "yuml", 255 },
   { NULL, 0 }
 };

void htmlformatt::do_entity(const string &name)
 {
  if(name=="#133")
    text("...");
  else if(!name.empty() && name[0]=='#')
   {
    char ch=atoi(name.c_str()+1);
    if(ch!=0) text(ch);
   }
  else if(name=="euro")
    text("EUR");
  else if(name=="copy")
    text("(C)");
  else if(name=="reg")
    text("(R)");
  else if(name=="nbsp")
   {
    bool old_attr_preformat=attr_preformat;
    attr_preformat=TRUE;
    text(' ');
    attr_preformat=old_attr_preformat;
   }
  else if(name=="middot")
   {
    bool old_attr_ibm_charset=attr_ibm_charset;
    attr_ibm_charset=TRUE;
    text(249);
    attr_ibm_charset=old_attr_ibm_charset;
   }
  else
   {
    for(unsigned i=0; entity_table[i].name!=NULL; i++)
      if(name==entity_table[i].name)
       {
        text(entity_table[i].code);
        return;
       }

    text('&');
    text(name);
    text(';');
   }
 }

void htmlformatt::entity_text(const string &str)
 {
  for(unsigned i=0; i<str.size(); i++)
   {
    if(str[i]=='&')
     {
      string entity_name;

      for(i++; i<str.size(); i++)
       {
        if(str[i]==';' || str[i]=='&' || str[i]==' ' || str[i]=='<')
         {
          do_entity(entity_name);
          break;
         }

        entity_name+=str[i];
       }
     }
    else
      text(str[i]);
   }
 }

void htmlformatt::entity(istream &in)
 {
  char ch;
  string entity_name;

  while(in.read(&ch, 1))
   {
    switch(ch)
     {
     case ';':
      do_entity(entity_name);
      return;

     case '&': // fix for idiots who do not know how to end an entity
      do_entity(entity_name);
      entity(in);
      return;

     case ' ': // fix for idiots who do not know &amp;
      text('&');
      text(entity_name);
      text(' ');
      return;

     case '<': // fix for idiots who do not know &amp;
      text('&');
      text(entity_name);
      tag(in);
      return;

     default:
      entity_name+=ch;
     }
   }
 }

void htmlformatt::new_line()
 {
  vector<htmlchart> empty_line;

  if(lines.size()==0)
    lines.push_back(empty_line);
  else
   {
    if(!do_get_max_width)
     {
      if(!lines[line].empty() && lines[line].size()<=width)
       {
        if(lines[line][0].center || current_alignment==CENTER)
         {
          unsigned i, insert=(width-lines[line].size())/2;
          vector<htmlchart> tmp_line;
          htmlchart htmlchar(' ');
          for(i=0; i<insert; i++) tmp_line.push_back(htmlchar);
          for(i=0; i<lines[line].size(); i++) tmp_line.push_back(lines[line][i]);
          lines[line]=tmp_line;
         }
        else if(current_alignment==RIGHT)
         {
          unsigned i, line_size=lines[line].size();
          if(line_size>0 && lines[line][line_size-1].ch==' ') line_size--;
          line_size=min(width, line_size);
          unsigned insert=(width-line_size);
          vector<htmlchart> tmp_line;
          htmlchart htmlchar(' ');
          for(i=0; i<insert; i++) tmp_line.push_back(htmlchar);
          for(i=0; i<line_size; i++) tmp_line.push_back(lines[line][i]);
          lines[line]=tmp_line;
         }
        else if(current_alignment==BLOCK)
         {
         }
       }
     }
    else if(!lines[line].empty())
     {
      unsigned line_size=lines[line].size();
      if(lines[line].back().ch==' ') line_size--;
      max_width=max(max_width, line_size);
     }

    lines.push_back(empty_line);
    line++;
   }
 }

void htmlformatt::line_break()
 {
  flush_word();

  if(lines.size()==0)
    new_line();
  else if(!lines[line].empty())
    new_line();
 }

void htmlformatt::gen_htmlchar(htmlchart &htmlchar)
 {
  htmlchar.ibm_charset=attr_ibm_charset;
  htmlchar.islink=attr_link;
  htmlchar.bold=attr_bold;
  htmlchar.underline=attr_underline;
  htmlchar.italic=attr_italic;
  htmlchar.preformat=attr_preformat;
  htmlchar.center=attr_center;
  if(attr_link) htmlchar.link=links.size()-1;
 }

void htmlformatt::flush_word()
 {
  if(word.empty()) return;

  if(attr_script)
   {
   }
  else if(attr_title)
   {
    if(title!="") title+=' ';
    for(unsigned i=0; i<word.size(); i++)
     {
      char ch=word[i].ch;
      if(word[i].ibm_charset) ibm2iso(&ch, 1);
      title+=ch;
     }
   }
  else
   {
    vector<htmlchart> empty_line;

    if(lines.size()==0)
      lines.push_back(empty_line);

    if((word.size()+lines[line].size())>width &&
       !lines[line].empty())
      new_line();

    if(lines[line].empty() && indent!=0)
     {
      htmlchart htmlchar(' ');
      for(unsigned i=0; i<indent; i++)
        lines[line].push_back(htmlchar);
     }

    for(unsigned i=0; i<word.size(); i++)
     {
      if(lines[line].size()>=width) 
       {
        new_line();
        if(indent!=0)
         {
          htmlchart htmlchar(' ');
          for(unsigned j=0; j<indent; j++)
            lines[line].push_back(htmlchar);
         }
       }
      lines[line].push_back(word[i]);
     }

    if(!lines[line].empty() && lines[line].size()<width)
     {
      htmlchart htmlchar(' ');
      gen_htmlchar(htmlchar);
      lines[line].push_back(htmlchar);
     }
   }

  word.clear();
 }

void htmlformatt::text(const string &str)
 {
  for(unsigned i=0; i<str.size(); i++)
    text(str[i]);
 }

void htmlformatt::text(char ch)
 {
  if(attr_preformat)
   {
    if(ch=='\r') return;
    if(ch=='\n') { flush_word(); new_line(); return; }
   }

  if(iscntrl(ch)) ch=' ';

  if(ch==' ' && !attr_preformat)
    flush_word();
  else
   {
    htmlchart htmlchar;
    gen_htmlchar(htmlchar);
    htmlchar.ch=ch;
    word.push_back(htmlchar);
   }
 }

void htmlformatt::format(istream &in)
 {
  char ch;

  line=0;
  while(in.read(&ch, 1))
   {
    switch(ch)
     {
     case '<': tag(in); break;
     case '&': entity(in); break;
     default: text(ch);
     }

    if(do_get_max_width && max_width>=width) return;

    if(lines.size()>MAX_LINES)
      break;
   }

  line_break();
  if(lines[line].empty()) lines.pop_back();
 }

htmlformatt::htmlformatt()
 {
  width=80;
  do_get_max_width=FALSE;
  default_alignment=LEFT;
  current_alignment=LEFT;
  indent=0;
  attr_bold=attr_underline=attr_title=attr_link=attr_table=
  attr_preformat=attr_italic=attr_ibm_charset=attr_center=
  attr_script=FALSE;
  table.clear();
 }

unsigned htmlformatt::get_max_width(istream &in)
 {
  max_width=0;
  do_get_max_width=TRUE;
  format(in);
  return max_width;
 }
