#ifndef DBOX_HTMLFORM_H

#define DBOX_HTMLFORM_H

#include <libdbox.h>
#include <vector.h>
#include <string>
#include <iostream.h>

#define MAX_LINES 10000

class htmlchart
 {
 public:
  bool ibm_charset, islink, bold, underline, italic, preformat, center;
  bool marked;
  char ch;
  unsigned link, form;
    
  void clear();
  htmlchart() { clear(); }
  htmlchart(char _ch) { clear(); ch=_ch; }
 };
 
struct tag_paramt
 {
  string id, value;
 };

const string &search(const vector<tag_paramt> &params, const string &id);
bool search(const vector<tag_paramt> &params, const string &id, 
            string &value);

class htmlformatt
 {
 public:
  string title; // ISO
 
  enum alignmentt { LEFT, CENTER, RIGHT, BLOCK }
    default_alignment, current_alignment;
    
  struct form_inputt
   {
    enum form_typet { HIDDEN, DEFAULT, SUBMIT } type;
    string name, value;
    unsigned size, posx, posy, scrollx, scrolly, sizex, sizey;
    unsigned form;
   };

  struct formt
   {
    string action, method;
   };
  
  class linkt
   {
   public:
    string href;
    
    bool is_form;
    unsigned form_input;
    
    linkt() { is_form=FALSE; form_input=0; }
   };
   
  vector<formt> forms;
  vector<form_inputt> form_inputs;
  vector<linkt> links;
 
  vector<vector <htmlchart> > lines;
  
  void format(istream &in);
  void set_width(unsigned _width) { width=_width; }
 
  htmlformatt();
  
 protected:
  unsigned width, line;
  unsigned indent, max_width;
  vector<int> indent_stack;
  bool attr_bold, attr_underline, attr_title, attr_link,
       attr_table, attr_italic, attr_preformat, attr_ibm_charset,
       attr_center, attr_script, do_get_max_width;
  vector <htmlchart> word;

  void gen_htmlchar(htmlchart &htmlchar);
  void new_line();
  void line_break();
  void flush_word();
  void tag_params(istream &in, vector <tag_paramt> &params);
  void do_tag(istream &in, const string &name, const vector <tag_paramt> &params);
  void get_tag(istream &in, string &name, vector <tag_paramt> &params);
  void do_entity(const string &name);
  unsigned get_max_width(istream &in);
  void tag(istream &in);
  void entity(istream &in);
  void html_comment(istream &in);
  void text(char ch);
  void text(const string &str);
  void entity_text(const string &str);
  
  // table support
  class tablet
   {
   public:
    bool border;
    unsigned columns;
    void clear();
    void set_params(const vector<tag_paramt> &params);
    
    struct columnt
     {
      string html;
      unsigned max_width, width, col_span;
      htmlformatt *hf;
      unsigned link_offset;
      alignmentt align;
      bool th;
     };
     
    class rowt
     {
     public:
      vector<columnt> columns;
      unsigned height;
      rowt() { height=0; }
     };
     
    vector<rowt> rows;
   } table;
 
  void render_table();
  void do_td(istream &in, const vector <tag_paramt> &params, bool th);
  void do_tr(const vector <tag_paramt> &params);
  void draw_table_border(bool upper);
  void draw_table_separator(unsigned row_nr);
  void draw_td(tablet::columnt &column, unsigned l);
  void td_merge(tablet::columnt &column);
  void form(const vector <tag_paramt> &params);
  void form_input(const vector <tag_paramt> &params);
 };
 
#endif
