/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <ctype.h>

#include <strstream.h>

#include <libdbox.h>
#include <charsettools.h>
#include <stringtools.h>

#include "htmlformat.h"

void htmlformatt::tablet::clear()
 {
  border=FALSE;
  columns=0;
  rows.clear();
 }

void htmlformatt::tablet::set_params(const vector<tag_paramt> &params)
 {
  string value;
  if(search(params, "border", value))
    border=FALSE;
  else
   {
    border=(value!="0");
   }
 }

void htmlformatt::do_td(istream &in, const vector<tag_paramt> &params, 
                        bool th)
 {
  flush_word();

  if(table.rows.empty() || !attr_table) return;

   {
    tablet::columnt empty_column;
    table.rows.back().columns.push_back(empty_column);
   }

  table.columns=max(table.columns, table.rows.back().columns.size());

  tablet::columnt &column=table.rows.back().columns.back();

  column.col_span=atoi(search(params, "colspan").c_str());
  if(column.col_span==0) column.col_span=1;

  column.th=th;

   {
    string align;
    search(params, "align", align);
    if(stringcasecmp(align, "center")==0)
      column.align=CENTER;
    else if(stringcasecmp(align, "right")==0)
      column.align=RIGHT;
    else if(stringcasecmp(align, "left")==0)
      column.align=LEFT;
    else
      column.align=th?CENTER:LEFT;
   }

  unsigned nested_table=0;

  char ch;
  while(in.read(&ch, 1))
   {
    switch(ch)
     {
     case '<':
       {
        string tag_name;
        bool add_tag=FALSE;
        vector<tag_paramt> tag_params;
        get_tag(in, tag_name, tag_params);

        if(nested_table==0)
         {
          if((stringcasecmp(tag_name, "/td")==0 ||
             stringcasecmp(tag_name, "/tr")==0))
            return;
          else if(stringcasecmp(tag_name, "/table")==0 ||
                  stringcasecmp(tag_name, "td")==0 ||
                  stringcasecmp(tag_name, "th")==0 ||
                  stringcasecmp(tag_name, "tr")==0)
           {
            do_tag(in, tag_name, tag_params);
            return;
           }
          else
           {
            if(stringcasecmp(tag_name, "table")==0) 
              nested_table++;
 
            add_tag=TRUE;
           }
         }
        else // nested
         {
          add_tag=TRUE;

          if(stringcasecmp(tag_name, "/table")==0)
            nested_table--;
         }

        if(add_tag && tag_name!="!--")
         {
          column.html+='<';
          column.html+=tag_name;
          for(unsigned i=0; i<tag_params.size(); i++)
           {
            column.html+=' ';
            column.html+=tag_params[i].id+'='+'"';
            column.html+=tag_params[i].value+'"';
           }
          column.html+='>';
         }
       }
      break;

     default: column.html+=ch;
     }
   }
 }

void htmlformatt::do_tr(const vector<tag_paramt> &params)
 {
  line_break();

  if(!attr_table) return;
  tablet::rowt empty_row;
  table.rows.push_back(empty_row);
 }

void htmlformatt::draw_table_border(bool upper)
 {
  if(!table.border) return;

  htmlchart htmlchar;
  htmlchar.ibm_charset=TRUE;
  htmlchar.center=attr_center;

  htmlchar.ch=upper?218:192;
  lines[line].push_back(htmlchar);

  tablet::rowt &row=
    upper?(table.rows.front()):(table.rows.back());

  for(unsigned column=0; column<row.columns.size(); column++)
   {
    if(column!=0)
     {
      htmlchar.ch=upper?194:193;
      lines[line].push_back(htmlchar);
     }

    htmlchar.ch=196;
    for(unsigned i=0; i<row.columns[column].width; i++)
      lines[line].push_back(htmlchar);
   }
     
  htmlchar.ch=upper?191:217;
  lines[line].push_back(htmlchar);

  new_line();
 }

void htmlformatt::draw_table_separator(unsigned row_nr)
 {
  if(!table.border) return;

  htmlchart htmlchar;
  htmlchar.ibm_charset=TRUE;
  htmlchar.center=attr_center;

  htmlchar.ch=195;
  lines[line].push_back(htmlchar);

  vector<htmlchart> line1, line2;

   {
    tablet::rowt &row=table.rows[row_nr-1];

    for(unsigned column=0; column<row.columns.size(); column++)
     {
      if(column!=0)
       {
        htmlchar.ch=193;
        line1.push_back(htmlchar);
       }

      htmlchar.ch=196;
      for(unsigned i=0; i<row.columns[column].width; i++)
        line1.push_back(htmlchar);
     }
   }
     
   {
    tablet::rowt &row=table.rows[row_nr];

    for(unsigned column=0; column<row.columns.size(); column++)
     {
      if(column!=0)
       {
        htmlchar.ch=194;
        line2.push_back(htmlchar);
       }

      htmlchar.ch=196;
      for(unsigned i=0; i<row.columns[column].width; i++)
        line2.push_back(htmlchar);
     }
   }

  for(unsigned i=0; i<max(line1.size(), line2.size()); i++)
   {
    char ch1, ch2, ch;
    if(i<line1.size()) ch1=line1[i].ch; else ch1=196;
    if(i<line2.size()) ch2=line2[i].ch; else ch2=196;

    if(ch1==(char)196 && ch2==(char)196)
      ch=196;
    else if(ch1!=(char)196 && ch2==(char)196) 
      ch=ch1;
    else if(ch1==(char)196)
      ch=ch2;
    else 
      ch=197;

    htmlchar.ch=ch;
    lines[line].push_back(htmlchar);
   }
     
  htmlchar.ch=180;
  lines[line].push_back(htmlchar);

  new_line();
 }

void htmlformatt::draw_td(tablet::columnt &column, unsigned l)
 {
  for(unsigned i=0; i<column.width; i++)
   {
    htmlchart htmlchar(' ');

    if(column.hf->lines.size()>l)
     {
      if(column.hf->lines[l].size()>i)
        htmlchar=column.hf->lines[l][i];
     }

    htmlchar.center=attr_center;
    if(htmlchar.islink) htmlchar.link+=column.link_offset;
    lines[line].push_back(htmlchar);
   }
 }

void htmlformatt::td_merge(tablet::columnt &column)
 {
  column.link_offset=links.size();

  unsigned i;

  for(i=0; i<column.hf->links.size(); i++)
    links.push_back(column.hf->links[i]);

  for(i=form_inputs.size(); 
      i<column.hf->form_inputs.size(); i++)
    form_inputs.push_back(column.hf->form_inputs[i]);

  for(i=forms.size(); 
      i<column.hf->forms.size(); i++)
    forms.push_back(column.hf->forms[i]);
 }

void htmlformatt::render_table()
 {
  unsigned row, column;

  line_break();

  if(!attr_table) return;
  attr_table=FALSE;

  if(table.rows.empty()) return;
  if(table.rows[0].columns.empty()) return;

  //table.border=TRUE;

  // 
  // initialize vectors
  //
  vector<unsigned> col_max_width, col_width;

  for(column=0; column<table.columns; column++)
   {
    col_max_width.push_back(0);
    col_width.push_back(0);
   }

  // 
  // find out max_width of columns
  // 
  
  for(row=0; row<table.rows.size(); row++)
    for(column=0; column<table.rows[row].columns.size(); column++)
     {
      istrstream td_in(table.rows[row].columns[column].html.c_str(), 
                       table.rows[row].columns[column].html.size());
      htmlformatt td_format;
      td_format.forms=forms;
      td_format.form_inputs=form_inputs;
      td_format.width=width;
      unsigned w=td_format.get_max_width(td_in);
      table.rows[row].columns[column].max_width=w;

      if(table.rows[row].columns[column].col_span<=1)
        col_max_width[column]=max(col_max_width[column], w);
      //else
      //  col_max_width[column]=max(col_max_width[column], 
      //    max(w/table.rows[row].columns[column].col_span, (unsigned)1));
     }

  // 
  // set column width
  // 

   {
    unsigned column_width=width/table.columns-1;

    unsigned total_width=0;

    if(table.border)
      total_width+=1+table.columns;
    else
      total_width+=table.columns-1;
      
    for(column=0; column<table.columns; column++)
     {
      col_width[column]=min(column_width, col_max_width[column]);
      total_width+=col_width[column];
     }

    while(total_width<width)
     {
      bool did_something=FALSE;

      for(column=0; column<table.columns; column++)
       {
        if(col_width[column]<col_max_width[column] &&
           total_width<width)
         {
          col_width[column]++;
          total_width++;
          did_something=TRUE;
         }
       }

      if(!did_something) break;
     }
   }

  // 
  // format columns
  // 

  for(row=0; row<table.rows.size(); row++)
    for(column=0; column<table.rows[row].columns.size(); column++)
     {
      tablet::columnt &col=table.rows[row].columns[column];

      col.width=0;
      for(unsigned i=0; i<col.col_span; i++)
        if(column+i<col_width.size())
         {
          if(i!=0) col.width+=1;
          col.width+=col_width[column+i];
         }

      if(col.width!=0)
       {
        htmlformatt *hf=new htmlformatt;
        istrstream td_in(col.html.c_str(), col.html.size());
        hf->width=col.width;
        hf->attr_bold=col.th;
        hf->default_alignment=col.align;
        hf->current_alignment=col.align;
        hf->forms=forms;
        hf->form_inputs=form_inputs;
        hf->format(td_in);
        col.hf=hf;
        table.rows[row].height=
          max(table.rows[row].height, hf->lines.size());
        td_merge(col);
       }
      else
        col.hf=(htmlformatt *)NULL;
     }

  // 
  // draw upper border
  // 
  draw_table_border(TRUE);

  // 
  // draw table
  // 

  for(row=0; row<table.rows.size(); row++)
   {
    if(row!=0) draw_table_separator(row);

    for(unsigned l=0; l<table.rows[row].height; l++)
     {
      htmlchart htmlchar;
      htmlchar.ibm_charset=TRUE;
      htmlchar.center=attr_center;

      if(table.border)
       {
        htmlchar.ch=179;
        lines[line].push_back(htmlchar);
       }

      for(column=0; column<table.rows[row].columns.size(); column++)
       {
        if(column!=0)
         {
          htmlchar.ch=table.border?179:' ';
          lines[line].push_back(htmlchar);
         }

        draw_td(table.rows[row].columns[column], l);
       }

      if(table.border)
       {
        htmlchar.ch=179;
        lines[line].push_back(htmlchar);
       }

      new_line();
     }
   }

  // 
  // draw lower border
  // 
  draw_table_border(FALSE);

  // 
  // clean up
  // 

  for(row=0; row<table.rows.size(); row++)
    for(column=0; column<table.rows[row].columns.size(); column++)
      if(table.rows[row].columns[column].hf!=(htmlformatt *)NULL)
        delete table.rows[row].columns[column].hf;
 }

