/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include "htmlview.h"

#include <ui.h>
#include <nls.h>
#include "help.h"
#include "online.h"
#include "url.h"

htmlviewt::~htmlviewt()
 {
 }

htmlviewt::htmlviewt()
 {
  attributes_set=FALSE;
  current_link=page=0;
  page_has_link=FALSE;
 }

void htmlviewt::search()
 {
  string text=search_text;

  while(TRUE)
   {
    if(status.dologout) return;

    online.setattr(2, FALSE);
    online.box(online.sizex/2-30, 10, 
               online.sizex/2+30, 16, 
               BOXFRAMEDBL, nls["online.search"], TRUE);

    online.setcursor(online.sizex/2-28, 12);
    online.printf("Text:");

    menut menu(&online);
    menu.x=online.sizex/2-30;
    menu.y=10;

    editstringmet editme;
    editme.id=0;
    editme.x=8;
    editme.y=2;
    editme.sizex=49;
    editme.maxlen=TEXTLEN-1;
    editme.text=&text;
    menu.add(&editme);

    ok_buttonmet ok_me;
    ok_me.id=1;
    ok_me.x=14;
    ok_me.y=4;
    menu.add(&ok_me);

    abort_buttonmet abort_me;
    abort_me.id=3;
    abort_me.x=38;
    abort_me.y=4;
    menu.add(&abort_me);

    for(bool startover=FALSE; !startover;)
     {
      switch(menu.menu())
       {
       case 1:
        if(status.dologout) break;
        search_text=text;
        search_next(TRUE, TRUE);
        return;

       case MENU_HELP:
        displayhelp("internet/www/search.html");
        redraw_page();
        if(status.dologout) break;
        startover=TRUE;
        break;

       default:
        redraw_page();
        return;
       }
     }
   }
 }

void htmlviewt::search_next(bool force_redraw, bool reset_search)
 {
  unsigned old_search_line=search_line;

  if(reset_search)
    search_line=0;
  else
    search_line++;

  if(search_text=="")
   {
    if(force_redraw) redraw_page();
    return;
   }

  while(TRUE)
   {
    if(search_line<html->lines.size())
     {
      for(unsigned i=0; i<html->lines[search_line].size(); i++)
       {
        unsigned j;
        for(j=0; j<search_text.size(); j++)
          if(tolower(search_text[j])!=
             tolower(html->lines[search_line][i+j].ch))
            break;
 
        if(j==search_text.size())
         {
          if(search_marked && !reset_search && old_search_line==search_line)
           {
            if(force_redraw) redraw_page();
            return;
           }

          for(unsigned k=0; k<search_text.size(); k++)
            html->lines[search_line][i+k].marked=TRUE;

          if(search_marked)
           {
            for(unsigned k=0; k<html->lines[old_search_line].size(); k++)
              html->lines[old_search_line][k].marked=FALSE;
           }

          unsigned new_page=search_line/lines_per_page;
          if(!force_redraw && page==new_page)
           {
            if(search_marked) draw_line(old_search_line);
            draw_line(search_line);
            reset_attr();
            set_cursor();
           }
          else
           {
            page=new_page;
            redraw_page();
           }

          search_marked=TRUE;
          return;
         }
       }

      search_line++;
     }
    else if(reset_search)
     {
      error_box(online, nls["online.searchnotfound"]);
      if(search_marked)
       {
        for(unsigned k=0; k<html->lines[old_search_line].size(); k++)
          html->lines[old_search_line][k].marked=TRUE;
        search_marked=FALSE;
       }

      redraw_page();
      return;
     }
    else
      search_line=0;
   }
 }

void htmlviewt::mark_link(unsigned link)
 {
  unsigned cursorx=0, cursory=0;
  reset_attr();

  unsigned start_line=page*lines_per_page;
  for(unsigned line=start_line; 
      line<(start_line+lines_per_page); line++)
   {
    if(line>=html->lines.size()) break;

    for(unsigned i=0; i<html->lines[line].size() &&
                      i<online.sizex; i++)
      if(html->lines[line][i].islink &&
         html->lines[line][i].link==link)
       {
        unsigned newx=1+i, newy=line+2-start_line;
        if(newx!=cursorx || newy!=cursory)
          online.setcursor(cursorx=newx, cursory=newy);
       
        set_attr(html->lines[line][i]);
        online.putch(html->lines[line][i].ch);
        cursorx++;
       }
   }

  online.setattr(1, FALSE);
  online.set_app_out_charset(IBM);
 }

bool htmlviewt::get_cursor_pos(unsigned &line,
                               unsigned &offs,
                               bool this_page)
 {
  line=0;
  offs=0;

  if(this_page && !page_has_link) return TRUE;

  unsigned start_line=0, end_line=html->lines.size();
  unsigned page_start=page*lines_per_page;
  unsigned page_end=page_start+lines_per_page;
  if(page_end>html->lines.size()) page_end=html->lines.size();

  unsigned posx=0;
  //, posy=0;
  if(html->links[current_link].is_form)
   {
    htmlformatt::form_inputt &form_input=
      html->form_inputs[html->links[current_link].form_input];
    posx=form_input.posx-form_input.scrollx;
   }
  else if(this_page)
   {
    start_line=page_start;
    end_line=page_end;
   }

  for(line=start_line; line<end_line; line++)
   {
    for(offs=0; offs<html->lines[line].size(); offs++)
      if(html->lines[line][offs].islink &&
         html->lines[line][offs].link==current_link)
       {
        if(posx==0)
         {
          if(this_page)
            return !(line>=page_start && line<page_end);

          return FALSE;
         }
        posx--;
       }
   }

  return TRUE;
 }

void htmlviewt::set_cursor()
 {
  unsigned line, offs;
  if(get_cursor_pos(line, offs, TRUE))
    online.setcursor(1, 1);
  else
    online.setcursor(1+offs, line+2-page*lines_per_page);
 }

bool htmlviewt::next_link(unsigned &link, bool first)
 {
  bool found_a_link=FALSE;
  unsigned new_link;

  for(unsigned line=page*lines_per_page; 
      line<((page+1)*lines_per_page); line++)
   {
    if(line>=html->lines.size()) break;

    for(unsigned i=0; i<html->lines[line].size(); i++)
      if(html->lines[line][i].islink)
       {
        unsigned l=html->lines[line][i].link;
        if(l>link || first)
         {
          if(found_a_link)
            new_link=min(new_link, l);
          else
           {
            found_a_link=TRUE;
            new_link=l;
           }
         }
       }
   }

  if(found_a_link) link=new_link;
  return found_a_link;
 }

bool htmlviewt::prev_link(unsigned &link, bool last)
 {
  bool found_a_link=FALSE;
  unsigned new_link;

  for(unsigned line=page*lines_per_page; 
      line<((page+1)*lines_per_page); line++)
   {
    if(line>=html->lines.size()) break;

    for(unsigned i=0; i<html->lines[line].size(); i++)
      if(html->lines[line][i].islink)
       {
        unsigned l=html->lines[line][i].link;
        if(l<link || last)
         {
          if(found_a_link)
            new_link=max(new_link, l);
          else
           {
            found_a_link=TRUE;
            new_link=l;
           }
         }
       }
   }

  if(found_a_link) link=new_link;
  return found_a_link;
 }

void htmlviewt::page_up()
 {
  if(page==0) return;

  page--;
  page_has_link=prev_link(current_link, TRUE);
  redraw_page();
 }

void htmlviewt::page_home()
 {
  if(page==0) return;

  page=0;
  page_has_link=next_link(current_link, TRUE);
  redraw_page();
 }

void htmlviewt::page_down()
 {
  if(page>=(pages-1)) return;

  page++;
  page_has_link=next_link(current_link, TRUE);
  redraw_page();
 }

void htmlviewt::page_end()
 {
  if(page>=(pages-1)) return;

  page=pages-1;
  page_has_link=prev_link(current_link, TRUE);
  redraw_page();
 }

void htmlviewt::link_up()
 {
  if(!page_has_link) { page_up(); return; }

  unsigned new_link=current_link;
  if(prev_link(new_link, FALSE))
   {
    unsigned old_link=current_link;
    current_link=new_link;
    mark_link(old_link);
    mark_link(current_link);
    set_cursor();
   }
  else
    page_up();
 }

void htmlviewt::link_down()
 {
  if(!page_has_link) { page_down(); return; }

  unsigned new_link=current_link;
  if(next_link(new_link, FALSE))
   {
    unsigned old_link=current_link;
    current_link=new_link;
    mark_link(old_link);
    mark_link(current_link);
    set_cursor();
   }
  else
    page_down();   
 }

void htmlviewt::follow_link(const string &link, string &new_url)
 {
  string protocol, host, path, port;
  const char *turl;

  turl=link.c_str();
  while(isalpha(*turl) || isdigit(*turl)) turl++;

  if(*turl==':')
   {
    parse_url(link, protocol, host, port, path);
    make_url(new_url, protocol, host, port, path);
    return;
   }

  if(link.empty() || link[0]=='#')
    return;

  parse_url(url, protocol, host, port, path);

  if(link[0]=='/')
    path=link.c_str()+1;
  else
   {
    turl=link.c_str();
    char new_path[DIRLEN*2];
    strmaxcpy(new_path, path.c_str(), DIRLEN-1);
 
    while(strncmp(turl, "../", 3)==0)
     {
      turl+=3;
     doagain:
      char *tptr=strrchr(new_path, '/');
      if(tptr!=NULL)
       {
        if(tptr[1]==0)
         {
          *tptr=0;
          goto doagain;
         }

        tptr[1]=0;
       }
     }

    if(new_path[strlen(new_path)-1]=='/')
      strcat(new_path, turl);
    else
     {
      char *tptr=strrchr(new_path, '/');

      if(stricmp(turl, "./")==0) turl="";

      if(tptr==NULL)
        strmaxcpy(new_path, turl, DIRLEN-1);
      else
        strmaxcpy(tptr+1, turl, TEXTLEN-1);
     }

    path=new_path;
   }

  make_url(new_url, protocol, host, port, path);
 }

bool htmlviewt::form_key(word key)
 {
  htmlformatt::form_inputt &form_input=
    html->form_inputs[html->links[current_link].form_input];
  char ch=lo(key);

  unsigned line, offs;
  bool cursor=get_cursor_pos(line, offs, TRUE);  

  switch(form_input.type)
   {
   case htmlformatt::form_inputt::DEFAULT:
    switch(key)
     {
     case Key_Lft:
      if(form_input.posx!=0 && cursor)
       {
        form_input.posx--;
        if(form_input.posx<form_input.scrollx)
         {
          form_input.scrollx--;
         }

        set_cursor(); 
       }
      return FALSE;

     case Key_Rgt:
      if(form_input.posx<form_input.value.size() && cursor)
       {
        form_input.posx++;
        if(form_input.posx-form_input.scrollx>form_input.sizex)
         {
          form_input.scrollx++;
         }

        set_cursor(); 
       }
      return FALSE;

     default:
      if(!iscntrl(ch))
       {     
        form_input.value.insert(form_input.posx, 1, ch);
        
        return FALSE;
       }
     }
    break;

   case htmlformatt::form_inputt::SUBMIT:
   case htmlformatt::form_inputt::HIDDEN:
    break;
   }

  return TRUE;
 }

bool htmlviewt::view(htmlformatt &_html, string &_url)
 {
  html=&_html;
  url=_url;

  search_text="";
  search_line=0;
  search_marked=FALSE;

  lines_per_page=online.sizey-2;

  internal=strnicmp(url.c_str(), "internal:", 9)==0;

  pages=html->lines.size()/lines_per_page;
  if((html->lines.size()%lines_per_page)!=0) pages++;
  if(pages==0) pages=1;
  if(page>=pages) page=pages-1;

  title=html->title;
  if(title=="") title=internal?"no title":url;

  if(!page_has_link)
    page_has_link=next_link(current_link, TRUE);
  redraw_page();

  while(TRUE)
   {
    word key=online.doreadkey();

    if(page_has_link && html->links[current_link].is_form)
     {
      if(!form_key(key)) continue;
     }

    switch(key)
     {
     case 0xffff: return TRUE;

     case Key_C_C:
     case Key_Esc: return TRUE;

     case Key_C_R: 
      redraw_page();
      break;

     case Key_F2:
      if(internal || 
         strnicmp(url.c_str(), "source:", 7)==0)
        break;

      _url="SOURCE";
      return FALSE;

     case Key_BS:
     case Key_Del:
      if(internal) break;
      _url="HISTORY";
      return FALSE;

     case Key_PgUp: page_up(); break;
     case Key_PgDn: page_down(); break;
     case Key_Lft: _url="BACK"; return FALSE;
     case Key_Up: link_up(); break;
     case Key_Dwn: link_down(); break;
     case Key_Home: page_home(); break;
     case Key_End: page_end(); break;
     case Key_F1: _url="HELP"; return FALSE;

     case Key_F3:
     case Key_C_F:
      search();
      break;

     case Key_Ent:
     case Key_Rgt:
      if(!page_has_link) break;
      follow_link(html->links[current_link].href, _url);
      return FALSE;

     default:
      switch(lo(key))
       {
       case '/':
        search();
        break;

       case 'v':
        if(internal) break;
        _url="BOOKMARKS";
        return FALSE;

       case 'a':
        if(internal) break;
        _url="ADD_BOOKMARK";
        return FALSE;

       case 'n':
        search_next(FALSE, FALSE);
        break;

       case '?':
        _url="HELP";
        return FALSE;

       case '=':
        if(!page_has_link) break;
        info_box(online, html->links[current_link].href);
        if(status.dologout) return TRUE;
        redraw_page();
        break;
 
       case 'h':
        if(internal) break;
        _url="HOME";
        return FALSE;

       case 'd':
        break;

       case 's':
        if(internal ||
           strnicmp(url.c_str(), "source:", 7)==0)
          break;

        _url="SOURCE";
        return FALSE;

       case 'u':
       case 'g':
        if(internal) break;
        if(!enter_url(_url)) return FALSE;
        if(status.dologout) return TRUE;
        redraw_page();
        break;
       }
     }
   }
  return TRUE;
 }

void htmlviewt::draw_statusline(const string &text)
 {
  int y, i, j;
  online.setattr(10, FALSE);
  y=text.size();
  j=(online.sizex-y)/2;
  online.setcursor(1, online.sizey);
  for(i=0; i<j; i++) online.putch(' ');
  online.set_app_out_charset(ISO);
  online.printf("%s", text.c_str());
  j=online.sizex-j-y-1;
  for(i=0; i<j; i++) online.putch(' ');
  online.setattr(1, FALSE);
 }

void htmlviewt::draw_title()
 {
  char *title_text;
  unsigned max_size=online.sizex-2;

  title_text=strdup(title.c_str());
  if(strlen(title_text)>max_size) title_text[max_size]=0;

  online.set_app_out_charset(ISO);
  ::draw_title(online, title);
  online.set_app_out_charset(IBM);

  free(title_text);

  //if(getbalance(&status.userprofile).amount!=0 && !internal)
  // {
  //  online.setattr(2, FALSE);
  //  printkonto();
  //  online.setattr(1, FALSE);
  // }
 }

void htmlviewt::set_attr(const htmlchart &htmlchar)
 {
  if(attributes_set)
   {
    if(htmlchar.islink==last_set_attr.islink &&
       htmlchar.bold==last_set_attr.bold &&
       htmlchar.italic==last_set_attr.italic &&
       htmlchar.ibm_charset==last_set_attr.ibm_charset)
      return;
   }

  if(htmlchar.marked)
   {
    online.setattr(2, htmlchar.bold);
   }
  else if(htmlchar.islink)
   {
    bool selected=(htmlchar.link==current_link);
    online.setattr(selected?11:10, htmlchar.bold);
   }
  else if(htmlchar.bold)
   {
    online.setattr(1, TRUE);
   }
  else if(htmlchar.italic)
   {
    if(online.usecolor)
      online.setcolor(3, 0);
    else
      online.setattr(1, TRUE);
   }
  else
    online.setattr(1, FALSE);

  if(htmlchar.ibm_charset)
    online.set_app_out_charset(IBM);
  else
    online.set_app_out_charset(ISO);

  last_set_attr=htmlchar;
 }

void htmlviewt::redraw_page()
 {
  if(status.dologout) return;
  draw_page();
  //if(page_has_link) mark_link(current_link);
  set_cursor();
 }

void htmlviewt::draw_page()
 {
  if(status.dologout) return;

  online.setattr(1, FALSE);
  online.clrscr();

  draw_title();

   {
    char text[TEXTLEN];
    if(internal ||
       strnicmp(url.c_str(), "error:", 6)==0 ||
       strnicmp(url.c_str(), "source:", 7)==0)
      snprintf(text, TEXTLEN, nls["online.pageinternal"], page+1, pages);
    else
      snprintf(text, TEXTLEN, nls["online.page"], 
        page+1, pages, url.c_str());
    text[77]=0;
    draw_statusline(text);
   }

  asyncio.flush_output();

  reset_attr();

  unsigned start_line=page*lines_per_page;
  for(unsigned line=start_line; 
      line<(start_line+lines_per_page); line++)
   {
    if(line>=html->lines.size()) break;
    draw_line(line);
   }

  online.setattr(1, FALSE);
  online.set_app_out_charset(IBM);
 }

void htmlviewt::draw_line(unsigned line)
 {
  if(line>=html->lines.size()) return;

  online.setcursor(1, line+2-(page*lines_per_page));

  for(unsigned i=0; i<html->lines[line].size() &&
      i<online.sizex; i++)
   {
    set_attr(html->lines[line][i]);
    online.printf("%c", (html->lines[line][i].ch));
   }
 }

void htmlviewt::reset_attr()
 {
  attributes_set=FALSE;
 }

bool htmlviewt::enter_url(string &new_url)
 {
  string _url=url;

  while(TRUE)
   {
    if(status.dologout) return TRUE;

    online.setattr(2, FALSE);
    online.box(online.sizex/2-30, 10, 
               online.sizex/2+30, 16, 
               BOXFRAMEDBL, nls["online.enterurl"], TRUE);


    menut menu(&online);
    menu.x=online.sizex/2-30;
    menu.y=10;

    online.setcursor(menu.x+2, menu.y+2);
    online.printf("URL:");

    editstringmet editme;
    editme.id=0;
    editme.x=7;
    editme.y=2;
    editme.sizex=50;
    editme.maxlen=TEXTLEN-1;
    editme.text=&_url;
    menu.add(&editme);

    buttonmet okme;
    okme.id=1;
    okme.x=14;
    okme.y=4;
    okme.selectkey=Key_C_Ent;
    okme.sizex=6;
    okme.text=nls["online.ok"];
    menu.add(&okme);

    buttonmet clearme;
    clearme.id=2;
    clearme.x=24;
    clearme.y=4;
    clearme.sizex=10;
    clearme.text=nls["online.newurl"];
    menu.add(&clearme);

    buttonmet abortme;
    abortme.id=3;
    abortme.x=38;
    abortme.y=4;
    abortme.sizex=9;
    abortme.text=nls["online.abort"];
    menu.add(&abortme);

    for(bool startover=FALSE; !startover;)
     {
      switch(menu.menu())
       {
       case 1:
        if(status.dologout) break;
        if(strnicmp(_url.c_str(), "http:", 5)!=0 &&
           strnicmp(_url.c_str(), "https:", 6)!=0 &&
           strnicmp(_url.c_str(), "finger:", 7)!=0 &&
           strnicmp(_url.c_str(), "news:", 5)!=0 &&
           strnicmp(_url.c_str(), "gopher:", 7)!=0 &&
           strnicmp(_url.c_str(), "ftp:", 4)!=0)
          new_url=string("http://")+_url;
        else 
          new_url=_url;
        return FALSE;

       case 2:
        editme.clear_text();
        menu.select(&editme);
        break;

       case MENU_HELP:
        displayhelp("internet/www/enterurl.html");
        if(status.dologout) break;
        redraw_page();
        if(status.dologout) break;
        startover=TRUE;
        break;

       default:
        return TRUE;
       }
     }
   }

  return TRUE;
 }

