/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <errno.h>
#include <netdb.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/wait.h>
#include <signal.h>

#include <logf.h>
#include <ui.h>
#include <nls.h>
#include "online.h"
#include "isockett.h"

isockett::isockett():sockett()
 {
 }

isockett::~isockett()
 {
 }

size_t isockett::read(void *buf, size_t count)
 {
  char *dest=(char *)buf;
  size_t bytes_read=0;

  while(bytes_read<count)
   {
    size_t result=sockett::read(dest, count-bytes_read);

    if(result>0)
     {
      dest+=result;
      bytes_read+=result;
     }
    else if(result==0)
      return bytes_read;

    if(!socket_open || status.dologout) return bytes_read;

    switch(online.readkey())
     {
     case 0xffff:
      return bytes_read;

     case Key_Esc:
      abort=TRUE;
      return bytes_read;
     }
   }

  return bytes_read;
 }

ibsockett::ibsockett():isockett()
 {
  progressbar_read=FALSE;
  progressbar_width=0;
  progressbar_size=0;
  progressbar_bytes=0;
 }

ibsockett::~ibsockett()
 {
 }

bool ibsockett::open(const string &_title, const string &address, word port)
 {
  title=_title;
  box_drawn=FALSE;
  time(&start);
  last_char=start;
  char_index=0;

  return isockett::open(address, port);
 }

void ibsockett::update_bar(bool force_draw)
 {
  if(!box_drawn) return;
  
  #define BAR_WIDTH 36

  byte width=0;

  if(progressbar_size==0)
    return;

  if(progressbar_bytes>=progressbar_size)
    width=BAR_WIDTH;
  else
    width=BAR_WIDTH*progressbar_bytes/progressbar_size;

  if(width==progressbar_width && !force_draw) return;
  progressbar_width=width;

  online.setcursor((online.sizex-BAR_WIDTH)/2, 10);
  online.setattr(11, FALSE);
  unsigned i;
  for(i=0; i<width; i++)
    online.printf(" ");
    
  online.setattr(1, FALSE);
  for(; i<BAR_WIDTH; i++)
    online.printf(" ");
  
  online.setattr(1, FALSE);
 }

size_t ibsockett::read(void *buf, size_t count)
 {
  char chars[]="|/-\\";

  char *dest=(char *)buf;
  size_t bytes_read=0;

  while(bytes_read<count)
   {
    size_t result=sockett::read(dest, count-bytes_read);

    if(result>0)
     {
      dest+=result;
      bytes_read+=result;
      progressbar_bytes+=result;
      update_bar(FALSE);
     }

    if(!socket_open || status.dologout) return bytes_read;

    if(!abort)
     {
      switch(online.readkey())
       {
       case 0xffff:
        return bytes_read;

       case Key_Esc:
        if(really_abort())
         {
          abort=TRUE;
          return bytes_read;
         }
   
        if(!status.dologout) draw_box();
       }

      time_t now;
      time(&now);

      if(box_drawn)
       {
        if(now>last_char)
         {
          online.setattr(2, FALSE);

          if(progressbar_size!=0)
           {
            char tempstr[TEXTLEN];
            snprintf(tempstr, TEXTLEN, "%u/%u Bytes",
              progressbar_bytes, progressbar_size);
            online.setcursor((online.sizex-strlen(tempstr))/2, 11);
            online.printf("%s", tempstr);
           }

          online.setcursor(online.sizex/2, 13);
          online.printf("%c", chars[char_index++]);
          asyncio.flush_output();
          if(chars[char_index]==0) char_index=0;
          last_char=now;
         }
       }
      else if(now>start)
       {
        draw_box();
        box_drawn=TRUE;
       }
     }
   }

  return bytes_read;
 }

void ibsockett::draw_box()
 {
  online.setattr(2, FALSE);

  online.box(online.sizex/2-20, 8, online.sizex/2+20, 18,
             BOXFRAMEDBL, title.c_str(), TRUE);

  const char *msg1=nls.get("online.pressESCforabort1");
  const char *msg2=nls.get("online.pressESCforabort2");

  online.set_app_out_charset(ISO);
  online.setcursor((online.sizex-strlen(msg1))/2, 15);
  online.printf("%s", msg1);
  online.setcursor((online.sizex-strlen(msg2))/2, 16);
  online.printf("%s", msg2);
  online.set_app_out_charset(IBM);

  online.setattr(1, FALSE);

  update_bar(TRUE);
 }

bool ibsockett::really_abort()
 {
  menut menu(&online);

  online.setattr(3, FALSE);

  online.box(online.sizex/2-20, 10, online.sizex/2+20, 16,
             BOXFRAMEDBL, NULL, TRUE);

  online.setcursor(online.sizex/2-11, 12);
  online.printf("Wirklich abbrechen?");

  buttonmet yesme;
  yesme.x=online.sizex/2-14;
  yesme.y=14;
  yesme.id=0;
  yesme.selectkey=Key_C_Ent;
  yesme.sizex=6;
  yesme.text="^Ja";
  menu.add(&yesme);

  buttonmet nome;
  nome.x=online.sizex/2+3;
  nome.y=14;
  nome.id=1;
  nome.sizex=6;
  nome.text="^Nein";
  menu.add(&nome);

  return menu.menu()==0;
 }

void ibsockett::set_progress_bar(bool read, size_t size)
 {
  progressbar_read=read;
  progressbar_width=BAR_WIDTH+1;
  progressbar_size=size;
  progressbar_bytes=0;
 }
