/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: 

*/

#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <errno.h>
#include <netdb.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/wait.h>
#include <signal.h>

#include <logf.h>
#include <asyncio.h>
#include <winport.h>
#include <configure.h>
#include "online.h"
#include "sockett.h"

int sockett::form(const char *format, ...)
 {
  va_list args;
  char buffer[2000];
  int chars;

  if(!socket_open) return 0;

  va_start(args, format);
  chars=vsnprintf(buffer, 2000, format, args);
  va_end(args);

  write(buffer, chars);
  return chars;
 }

char *sockett::gets(char *dest, size_t maxlen)
 {
  size_t len=0;
  char *orig_dest=dest;

  while(socket_open && !abort)
   {
    char ch;
    size_t i=read(&ch, 1);

    if(i==1)
     {
      if(len<(maxlen-1))
       {
        dest[0]=ch;
        dest++;
        len++;
       }

      if(ch=='\n')
       {
        dest[0]=0;
        return orig_dest;
       }
     }
   }
  
  return NULL;
 }

void sockett::get_socket_state()
 {
  int error=0;
  socklen_t optlen=sizeof(error);
  fd_set write_fd;
  timeval timeout;
  timeout.tv_usec=timeout.tv_sec=0;

  if(!socket_open) return;
  if(!connecting) return;

  FD_SET(fd, &write_fd);
  if(select(fd+1, NULL, &write_fd, NULL, &timeout)<0)
   {
    socket_open=FALSE;
    return;
   }

  if(FD_ISSET(fd, &write_fd))
   {
    getsockopt(fd, SOL_SOCKET, SO_ERROR, &error, &optlen);

    if(error==EINPROGRESS)
     {
     }
    else if(error==0)
     {
      connecting=FALSE;
      if(verbose) input_buffer+=" Ok\r\n";
     }
    else
     {
      socket_error=error;
      asyncio.drop_output(fd);
      socket_open=FALSE;

      if(verbose)
       {
        input_buffer+=" failed: "; 
        input_buffer+=strerror(socket_error);
        input_buffer+="\r\n";
       }
     } 
   }
 }

size_t sockett::read(void *buf, size_t count)
 {
  if(input_buffer.size()>0)
   {
    ssize_t copied=min(input_buffer.size(), count);
    memcpy(buf, input_buffer.c_str(), copied);
    input_buffer.erase(0, copied);
    return copied;
   }

  get_socket_state();
  if(!socket_open) return 0;

  if(connecting)
   {
    if(status.dologout) return 0;
    vector<int> handles;
    handles.push_back(fd);
    handles.push_back(online.console.fd);
    if(status.online2)
      handles.push_back(online.comport.fd);
    asyncio.input(handles);
    return 0;
   }

  int loops=2;

  do
   {
    ssize_t copied=0;

    if(count==1)
     {
      #define BUFSIZE 100
      char buffer[BUFSIZE];
      copied=::read(fd, buffer, BUFSIZE);
      if(copied>0)
       {
        input_buffer.assign(buffer+1, copied-1);
        ((char *)buf)[0]=buffer[0];
        return 1;
       }
     }
    else
     {
      copied=::read(fd, buf, count);
      if(copied>0) return copied;
     }

    if(copied==0)
     {
      socket_error=0;
      socket_open=FALSE;
      return 0;
     }

    // here: copied<0

    if(errno!=EAGAIN)
     {
      socket_error=errno;
      asyncio.drop_output(fd);
      socket_open=FALSE;
      return 0;
     }

    if(status.dologout) return 0;

    vector<int> handles;
    handles.push_back(fd);
    handles.push_back(online.console.fd);
    if(status.online2)
      handles.push_back(online.comport.fd);

    if(asyncio.input(handles))
     {
      socket_error=errno;
      asyncio.drop_output(fd);
      socket_open=FALSE;
      return 0;
     }
   }
  while((loops--)>0);

  return 0;
 }

size_t sockett::write(const void *buf, size_t count)
 {
  if(!socket_open) return 0;
  get_socket_state();
  if(!socket_open) return 0;
  asyncio.write(fd, (const char *)buf, count);
  return count;
 }

bool sockett::open(const string &address, word port)
 {
  socket_error=0;

  if(verbose)
    input_buffer+="Looking up "+address+"...";

  struct sockaddr_in sock_out;
  struct hostent *hostent;

  hostent=gethostbyname(address.c_str());
  if(hostent==NULL)
   {
    if(verbose)
      input_buffer+=" host lookup failed\r\n";

    socket_error=h_errno;
    socket_error_type=LOOKUP_ERROR;
    return TRUE;
   }
  else if(verbose)
   {
    input_buffer+=" ";
    input_buffer+=inet_ntoa(*(in_addr *)hostent->h_addr);
    input_buffer+="\r\n";
   }

  fd=socket(AF_INET, SOCK_STREAM, 0);
  if(fd<0)
   {
    socket_error=errno;
    return TRUE;
   }

  asyncio.make_async(fd);

  memset(&sock_out, 0, sizeof(sock_out));
  memcpy(&sock_out.sin_addr, hostent->h_addr, 4);
  sock_out.sin_port=htons(port);
  sock_out.sin_family=PF_INET;

  if(verbose)
    input_buffer+="Connecting...";

  if(connect(fd, (struct sockaddr *)&sock_out, sizeof(sock_out))<0)
   {
    if(errno!=EINPROGRESS) 
     {
      socket_error=errno;
      ::close(fd);
      fd=-1;

      if(verbose)
       {
        input_buffer+="failed ";
        input_buffer+=strerror(socket_error);
        input_buffer+="\r\n";
       }

      return TRUE;
     } 
    else
      connecting=TRUE;
   }
  else if(verbose)
    input_buffer+=" Ok\r\n";

  socket_open=TRUE;
  abort=FALSE;

  return FALSE;
 }

bool sockett::close()
 {
  asyncio.flush_output();
  if(fd>=0) { ::close(fd); fd=-1; }
  socket_open=FALSE;
  return FALSE;
 }
 
