/*

  DBOX Module
  Copyright (C) 1994-2000 Daniel Kroening <kroening@handshake.de>

  Purpose: NNTP gateway for web browser

*/

#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>

#include <algorithm>
#include <set.h>
#include <strstream.h>
#include <fstream.h>

#include <config.h>
#include <logf.h>
#include <stringtools.h>
#include <dboxversion.h>
#include <mtoolrfc.h>
#include <htmltools.h>
#include <nls.h>
#include "online.h"
#include "url.h"
#include "htmlview.h"
#include "writemail.h"
#include "basic_www.h"
#include "formatmessage.h"

bool do_basic_wwwt::nntp_error(ibsockett &socket)
 {
  string new_url="error://";

  if(socket.get_error_type()==sockett::SOCKET_ERROR)
    new_url+="open.socket/";
  else
    new_url+="open.lookup/";

  new_url+=i2string(socket.get_error_number())+"/"+url;
  parse_url(new_url, protocol, host, port, path);
  return get_and_format_error();
 }

bool do_basic_wwwt::open_nntp(ibsockett &socket)
 {
  word port_nr=atol(port.c_str());
  if(port_nr==0) port_nr=119;

  if(socket.open("NNTP", host, port_nr))
    return TRUE;

  return FALSE;
 }

bool do_basic_wwwt::nntp_get_active()
 {
  ibsockett socket;

  if(news_active_cache_host==host) return FALSE;

  if(open_nntp(socket)) 
   {
    nntp_error(socket);
    return TRUE;
   }

  string request="MODE READER\r\nLIST ACTIVE\r\n";
  socket.write(request.c_str(), request.size());

  string message;

  char line[1000];
  unsigned nr=0;

  while(TRUE)
   {
    if(socket.gets(line, 999)==NULL)
     {
      string new_url="error://read/"+
                     i2string(socket.get_error_number())+"/"+url;
      parse_url(new_url, protocol, host, port, path);
      return get_and_format_error();
     }

    if(nr>=3)
     {
      if(line[0]!='.' || (line[1]!='\n' && line[1]!='\r' && line[1]!=0))
       {
        char *tptr=strchr(line, ' ');
        if(tptr!=NULL) *tptr=0;
        news_active.push_back(line);
       }
      else
        break;
     }
    else if(line[0]!='2')
     {
      string new_url="error://nntp/";
      for(unsigned i=0; line[i]!=0; i++) 
        if(!iscntrl(line[i])) new_url+=line[i];
      parse_url(new_url, protocol, host, port, path);
      return get_and_format_error(); 
     }

    nr++;
   }

  sort(news_active.begin(), news_active.end());

  news_active_cache_host=host;

  return FALSE;  
 }

bool do_basic_wwwt::get_and_format_nntp()
 {
  if(path=="" || path=="/")
    path="/*";

  if(strncmp(path.c_str(), "/<", 2)==0)
   {
    ibsockett socket;

    if(open_nntp(socket)) 
      return nntp_error(socket);

    string request="MODE READER\r\nARTICLE "+string(path.c_str()+1)+"\r\n";
    socket.write(request.c_str(), request.size());

    string message;

    char line[1000];
    unsigned nr=0;

    while(TRUE)
     {
      if(socket.gets(line, 999)==NULL)
       {
        string new_url="error://read/"+
                       i2string(socket.get_error_number())+"/"+url;
        parse_url(new_url, protocol, host, port, path);
        return get_and_format_error();
       }

      if(nr>=3)
       {
        if(line[0]!='.' || (line[1]!='\n' && line[1]!='\r' && line[1]!=0))
          message+=line;
        else
          break;
       }
      else if(line[0]!='2')
       {
        string new_url="error://nntp/";
        for(unsigned i=0; line[i]!=0; i++) 
          if(!iscntrl(line[i])) new_url+=line[i];
        parse_url(new_url, protocol, host, port, path);
        return get_and_format_error();
       }

      nr++;
     }

    source=message;
    rfcmessaget msg;
    msg.data=message.c_str();
    msg.endofmail=msg.endoffile=message.c_str()+message.size();
    msg.parse();

    mimestructt mimestruct;
    mimestruct.mimeparse(msg);

    string title="News Article";
    ostrstream out;
    out << "<html><head><title>" << title << "</title></head>\n"
        << "<body>\n";
    format_message(out, FALSE, FALSE, 0, mimestruct, msg);
    out << "</body></html>\n";
    format_html(out.str());
    out.freeze(0);
   }
  else
   {
    if(nntp_get_active()) return FALSE;

    set<string> newsgroups;
    for(vector<string>::const_iterator it=news_active.begin(); 
        it!=news_active.end(); it++)
     {
      string name;
      string::size_type pos=it->find('.');

      if(pos==string::npos)
        name=*it;
      else
       {
        name.assign(*it, 0, pos);
        name+=".*";
       }

      if(newsgroups.find(name)==newsgroups.end())
        newsgroups.insert(name);
     }

    string page="<html><head><title>Newsgroups</title></head><body><pre>\n";

    for(set<string>::const_iterator it=newsgroups.begin();
        it!=newsgroups.end(); it++)
     {
      page+="<a href=\"/";
      page+=htmlize_string(*it);
      page+="\">";
      page+=htmlize_string(*it);
      page+="</a>";
      page+="\n";
     }

    page+="</pre></body></html>\n";
    format_html(page);
   }

  return FALSE;
 }

