#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include <libdbox.h>
#include <asyncio.h>

asynciot asyncio;

io_channelt::io_channelt()
 {
  handle=-1;
 }

io_channelt::~io_channelt()
 {
 }

void asynciot::write(io_channelt &io_channel, const char *data, size_t len)
 {
  io_channel.buffer.append(data, len);

  if(io_channel.buffer.size()>200)
    do_output(io_channel);
 }

void asynciot::write(int handle, const char *data, size_t len)
 {
  for(unsigned i=0; i<channels->size(); i++)
    if((*channels)[i].handle==handle)
     {
      write((*channels)[i], data, len);
      return;
     }

  io_channelt io_channel;
  io_channel.handle=handle;
  channels->push_back(io_channel);
  write(channels->back(), data, len);
 }

bool asynciot::input(const vector<int> &handles)
 {
  fd_set readset, writeset, exceptionset;
  unsigned i;
  int max_handle=0;
  timeval timeout;

  timeout.tv_sec=0;
  timeout.tv_usec=100000;

  FD_ZERO(&readset);
  FD_ZERO(&writeset);
  FD_ZERO(&exceptionset);

  for(i=0; i<handles.size(); i++)
   {
    FD_SET(handles[i], &readset);
    FD_SET(handles[i], &exceptionset);
    max_handle=max(max_handle, handles[i]);
   }

  for(i=0; i<(*channels).size(); i++)
    if(!(*channels)[i].buffer.empty())
     {
      FD_SET((*channels)[i].handle, &writeset);
      max_handle=max(max_handle, (*channels)[i].handle);
     }

  if(select(max_handle+1, &readset, &writeset, &exceptionset, &timeout)<0)
    return TRUE;

  for(i=0; i<(*channels).size(); i++)
    if(!(*channels)[i].buffer.empty())
       // FD_ISSET((*channels)[i].handle, &writeset))
      do_output((*channels)[i]);

  return FALSE;
 }

void asynciot::do_output(io_channelt &io_channel)
 {
  ssize_t result;

  if(io_channel.buffer.empty()) return;

  if(io_channel.handle<0) abort();

  result=::write(io_channel.handle, 
                 io_channel.buffer.c_str(),
                 io_channel.buffer.size());

  if(result<0)
   {
   }
  else if(result==(signed)io_channel.buffer.size())
    io_channel.buffer="";
  else if(result!=0)
    io_channel.buffer.erase(0, result);
 }

void asynciot::flush_output()
 {
  time_t start;

  time(&start);

  while(TRUE)
   {
    bool flushed=TRUE;

    if(time(NULL)>start+3) 
     {
      drop_output();
      return;
     }

    for(unsigned i=0; i<channels->size(); i++)
      if(!(*channels)[i].buffer.empty())
        flushed=FALSE;

    if(flushed) return;

    vector<int> handles;
    input(handles);
   }
 }

bool asynciot::output_pending(int handle)
 {
  for(unsigned i=0; i<channels->size(); i++)
    if((*channels)[i].handle==handle)
      return !((*channels)[i].buffer.empty());

  return FALSE;
 }
  
bool asynciot::initialize()
 {
  channels=new vector<io_channelt>;
  if(channels==NULL) return TRUE;

  return FALSE;
 }

void asynciot::make_async(int handle)
 {
  fcntl(handle, F_SETFL, O_NONBLOCK);
 }

void asynciot::drop_output(int handle)
 {
  for(unsigned i=0; i<channels->size(); i++)
    if((*channels)[i].handle==handle)
      (*channels)[i].buffer="";
 }

void asynciot::drop_output()
 {
  for(unsigned i=0; i<channels->size(); i++)
    (*channels)[i].buffer="";
 }

void asynciot::clear()
 {
  if(channels!=NULL)
   {
    delete channels;
    channels=(vector<io_channelt> *)NULL;
   }
 }
