#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <libdbox.h>
#include <cmdline.h>
#include <logf.h>

cmdlinet::cmdlinet()
 {
  argcount=count=0;
 }

cmdlinet::~cmdlinet()
 {
  clear();
 }

void cmdlinet::clear()
 {
  int i;

  for(i=0; i<count; i++)
    if(options[i].optstring!=NULL)
      free(options[i].optstring);

  argcount=count=0;
 }

bool cmdlinet::isset(char option)
 {
  int i;

  i=getoptnr(option);
  if(i<0) return FALSE;
  return options[i].isset;
 }

bool cmdlinet::isset(const char *option)
 {
  int i;

  i=getoptnr(option);
  if(i<0) return FALSE;
  return options[i].isset;
 }

const char *cmdlinet::getval(char option)
 {
  int i;

  i=getoptnr(option);
  if(i<0) return (const char *)NULL;
  return options[i].value;
 }

const char *cmdlinet::getval(const char *option)
 {
  int i;

  i=getoptnr(option);
  if(i<0) return (const char *)NULL;
  return options[i].value;
 }

int cmdlinet::getoptnr(char option)
 {
  int i;

  for(i=0; i<count; i++)
    if(options[i].optchar==option)
      return i;
  
  return -1;
 }

int cmdlinet::getoptnr(const char *option)
 {
  int i;

  for(i=0; i<count; i++)
    if(options[i].optstring!=NULL && 
       strcmp(options[i].optstring, option)==0)
      return i;
  
  return -1;
 }

bool cmdlinet::parse(int argc, char **argv, const char *optstring)
 {
  int i, optnr;
  char tempstr[TEXTLEN];

  for(i=0; i<CMDLINE_MAXARGS; i++)
    args[i]=(char *)NULL;

  argcount=count=0;

  while(optstring[0]!=0)
   {
    if(optstring[0]==':')
      logf.internalerror("cmdlinet::parse", "Invalid option string");

    if(optstring[0]=='(')
     {
      options[count].islong=TRUE;
      options[count].optchar=0;
      options[count].isset=FALSE;
      options[count].value=(char *)NULL;
      optstring++;

      tempstr[0]=0;
      for(i=0; optstring[0]!=')' && optstring[0]!=0; i++)
       {
        if(i<(TEXTLEN-1)) { tempstr[i]=optstring[0]; tempstr[i+1]=0; }
        optstring++;
       }

      options[count].optstring=strdup(tempstr);
      if(optstring[0]==')') optstring++;
     }
    else
     {
      options[count].islong=FALSE;
      options[count].optchar=optstring[0];
      options[count].optstring=(char *)NULL;
      options[count].isset=FALSE;
      options[count].value=(char *)NULL;

      optstring++;
     }

    if(optstring[0]==':')
     {
      options[count].hasval=TRUE;
      optstring++;
     }
    else
      options[count].hasval=FALSE;

    count++;
   }

  for(i=1; i<argc; i++)
   {
    if(argv[i][0]!='-')
      args[argcount++]=argv[i];
    else
     {
      if(argv[i][1]=='-')
        optnr=getoptnr(argv[i]+2);
      else
        optnr=getoptnr(argv[i][1]);
   
      if(optnr<0) return TRUE;
      options[optnr].isset=TRUE;
      if(options[optnr].hasval)
       {
        if(argv[i][2]==0 || options[optnr].islong)
         {
          i++;
          if(i==argc) return TRUE;
          if(argv[i][0]=='-') return TRUE;
          options[optnr].value=argv[i];
         }
        else
          options[optnr].value=argv[i]+2;
       }
     }
   }

  return FALSE;
 }
