#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>

#include <tools.h>
#include <nls.h>
#include <charsettools.h>
#include <stringtools.h>
#include <mimedecode.h>
#include <rfc822.h>
#include <rfc1522.h>
#include <htmltools.h>
#include <formatmessage.h>
#include <parsefrom.h>

void get_icon(string &icon, const mimestructt &mimestruct)
 {
  if(mimestruct.content_type.value=="audio")
    icon="sound";
  else if(mimestruct.content_type.value=="video")
    icon="movie";
  else if(mimestruct.content_type.value=="text" ||
          mimestruct.content_type.value=="image")
    icon=mimestruct.content_type.value;
  else if(mimestruct.content_type.value=="application" &&
          mimestruct.content_type.subvalue=="postscript")
    icon="ps";
  else if(mimestruct.content_type.value=="application" &&
          mimestruct.content_type.subvalue=="pdf")
    icon="pdf";
  else
    icon="unknown";
 }

void displaymailstruct(ostream &out, const mimestructt &mimestruct)
 {
  string icon;

  get_icon(icon, mimestruct);

  out << "<img src=\"/html/icons/anthony-" << icon << ".gif\" "
         "border=0>";

  out << mimestruct.content_type.value << "/"
      << mimestruct.content_type.subvalue;

  for(unsigned j=0; j<mimestruct.content_type.attributes.size(); j++)
    out << "; " << mimestruct.content_type.attributes[j].name
        << "=\""  << mimestruct.content_type.attributes[j].value
        << '"';

  out << "\n";

  out << "<br>";
  out << mimestruct.content_transfer_encoding.value << "\n";

  if(mimestruct.sub.size()!=0)
   {
    out << "<ul>\n";

    for(unsigned i=0; i<mimestruct.sub.size(); i++)
     {
      out << "<li>\n";
      displaymailstruct(out, mimestruct.sub[i]);
     }

    out << "</ul>\n";
   }

  out << "<p>\n";
 }

void format_text_plain(ostream &out, const string &text)
 {
  out << "<code>";
  for(unsigned i=0; i<text.size(); i++)
   {
    char data=text[i];
    switch(data)
     {
     case ' ':
      if(i==0 ||
         (i<(text.size()-1) && text[i+1]==' ') ||
         (i>0 && (text[i-1]==' ' || text[i-1]=='\n')))
        out << "&nbsp;";
      else 
        out << ' ';
      break;

     case 'h':
     case 'f':
      if(strncmp(text.c_str()+i, "http://", 7)==0 ||
         strncmp(text.c_str()+i, "ftp://", 6)==0)
       {
        string url;

        for(; i<text.size(); i++)
          if(text[i]==' ' || text[i]=='\r' || text[i]=='\n' ||
             text[i]=='"' || text[i]=='<')
            break;
          else
            url+=text[i];

        out << "<a href=\"" << urlize_string(url) << "\">" 
            << htmlize_string(url) << "</a>";
       }
      else
        out << data;
      break;

     case 0: break;
     case '\r': break;
     case '\n': out << "<br>\n"; break;
     case '<': out << "&lt;"; break;
     case '>': out << "&gt;"; break;
     case '&': out << "&amp;"; break;
     default: out << data;
     }
   }
  out << "</code>\n";
 }

void displaymail(ostream &out, 
                 const string &msg_id,
                 const string &attachment,
                 const mimestructt &mimestruct,
                 const rfcmessaget &rfcm,
                 unsigned attachment_nr)
 {
  string attachment_type;

  bool do_inline=
    (mimestruct.content_disposition.value=="" ||
    stringcasecmp(mimestruct.content_disposition.value, "inline")==0);

  if(stringcasecmp(mimestruct.content_type.value, "text")==0)
   {
    if(stringcasecmp(mimestruct.content_type.subvalue, "plain")==0)
     {
      if(do_inline)
       {
        if(attachment_nr!=0) out << "\n<hr>\n";
        format_text_plain(out, mime_decode(mimestruct, rfcm));
       }
      else
        attachment_type="plain text";
     }
    else if(stringcasecmp(mimestruct.content_type.subvalue, "html")==0)
      attachment_type="html text";
    else
      attachment_type="text of unknown type";
   }
  else if(stringcasecmp(mimestruct.content_type.value, "message")==0)
   {
    if(stringcasecmp(mimestruct.content_type.subvalue, "rfc822")==0)
     {
      if(do_inline)
       {
        if(attachment_nr!=0) out << "\n<hr>\n";
        format_text_plain(out, mime_decode(mimestruct, rfcm));
       }
      else
        attachment_type="plain text";
     }
    else
      attachment_type="message of unknown type";
   }
  else if(stringcasecmp(mimestruct.content_type.value, "application")==0)
   {
    attachment_type="application data";
   }
  else if(stringcasecmp(mimestruct.content_type.value, "multipart")==0)
   {
    for(unsigned i=0; i<mimestruct.sub.size(); i++)
     {
      char nr[TEXTLEN];
      snprintf(nr, TEXTLEN, "%s%u", attachment==""?"":".", i+1);
      displaymail(out, msg_id, attachment+nr,
                  mimestruct.sub[i], rfcm, i);
     }
   }
  else
    attachment_type="unknown";

  if(attachment_type!="")
   {
    string name, icon, filename;

    mimestruct.content_type.findattribute(name, "name");

    mimestruct.content_disposition.findattribute(filename, "filename");
    if(filename=="") filename=name;
    filename=decode_rfc1522(filename);

    if(name=="")
      name=attachment_type;
    else
      name=decode_rfc1522(name);

    get_icon(icon, mimestruct);

    out << nls["webonline.attachment"]
        << ": <a href=\"/email";

    if(filename!="")
     {
      for(unsigned i=0; i<filename.size(); i++)
        if(filename[i]=='&' || filename[i]=='/' || filename[i]=='\\' ||
           filename[i]=='`' || isspace(filename[i]) || 
           iscntrl(filename[i]) || filename[i]=='\'')
          filename[i]='_';

      out << "/" << htmlize_string(filename);
     }

    out << "?action=read" << msg_id
        << "&attachment=" << attachment << "\">" 
           "<img src=\"/html/icons/anthony-" << icon << ".gif\" "
           "border=0><code>"
        << name
        << "</a></code>\n"
        << " (" << mimestruct.content_type.value << "/" <<
                   mimestruct.content_type.subvalue << ")"
        << "<p>\n";
   }
 }

void format_message(ostream &out,
                    bool show_mimestruct, bool show_source,
                    const string &msg_id,
                    const mimestructt &mimestruct,
                    const rfcmessaget &rfcm)
 {
  out << "<TABLE>\n";

   {
    string subject;
    rfcm.findsingleheader(subject, "subject");
    subject=htmlize_string(decode_rfc1522(subject));
    out << "<tr><th align=right valign=baseline>Betreff:</th>\n"
           "    <td valign=baseline><font size=5>" << subject
        << "</font></td></tr>\n";
   }

   {
    string date;
    rfcm.findsingleheader(date, "date");
    out << "<tr><th align=right>" << nls["webonline.date"] 
        << ":</th>\n<td>" << htmlize_string(date) << "</td>\n";
   }

   {
    string from, mailto, name;

    rfcm.findsingleheader(from, "from");
    parse_from(from, mailto, name);

    out << "<tr><th align=right>" << nls["webonline.from"]
        << ":<td><a href=\"mailto:" 
        << htmlize_string(mailto) << "\">"
        << htmlize_string(mailto) << "</a>";

    if(name!="")
      out << " (" << htmlize_string(decode_rfc1522(name)) << ")\n";
   }

   {
    string recipient;
    rfcm.findsingleheader(recipient, "X-POP3-Rcpt");
    out << "<tr><th align=right>" << nls["webonline.recipient"] 
        << ":</th>\n<td>" << htmlize_string(recipient) << "</td>\n";
   }

  out << "</TABLE><hr size=4><p>\n";

  if(show_mimestruct)
   {
    out << "Document structure:<p>\n";
    displaymailstruct(out, mimestruct);
   }
  else  
    displaymail(out, msg_id, "", mimestruct, rfcm, 0);
 }

