#include <ctype.h>

#include <mimedecode.h>
#include <stringtools.h>
#include <sbase64.h>

string mime_qp_decode(const string &bufcoded,
                      bool convert_underscore)
 {
  string bufplain;
  int state=0;
  char qp_char1=0, qp_char2=0;

  for(unsigned i=0; i<bufcoded.size(); i++)
   {
    char data=bufcoded[i];

    switch(state)
     {
     case 0:
      if(data=='=')
        state++;
      else
       {
        if(convert_underscore && data=='_')
          bufplain+=' ';
        else
          bufplain+=data;
       }
      break;

     case 1:
      if(data=='\r')
       {
       }
      else if(data=='\n')
        state=0;
      else if(isxdigit(data))
       {
        qp_char1=toupper(data);
        state++;
       }
      else
        state=0;
      break;
 
     case 2:
      if(isxdigit(data))
       {
        qp_char2=toupper(data);
        if(isdigit(qp_char1)) qp_char1-='0'; else qp_char1-='A'-10;
        if(isdigit(qp_char2)) qp_char2-='0'; else qp_char2-='A'-10;
        bufplain+=char((qp_char1<<4)|qp_char2);
       }
      state=0;
      break;
     }
   }  

  return bufplain;
 }

string mime_base64_decode(const string &bufcoded)
 {
  string filtered;

  for(unsigned i=0; i<bufcoded.size(); i++)
   {
    char data=bufcoded[i];
    switch(data)
     {
     case '\r':
     case '\n':
      break;
     default: filtered+=data;
     }
   }

  return sbase64_decode(filtered);
 }

string mime_decode(const mimestructt &mimestruct, const rfcmessaget &rfcm)
 {
  string source;

  source.assign(rfcm.data+mimestruct.offset+mimestruct.headerlen,
                mimestruct.len);

  if(stringcasecmp(mimestruct.content_transfer_encoding.value,
     "quoted-printable")==0)
   {
    return mime_qp_decode(source, FALSE);
   }
  else if(stringcasecmp(mimestruct.content_transfer_encoding.value, 
     "base64")==0)
   {
    return mime_base64_decode(source);
   }

  return source;
 }
