#include <stdio.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>
#include <termios.h>

#include <interlib.h>
#include <oldterm.h>
#include <charsettools.h>
#include <logf.h>
#include <keys.h>

char *charsets[]=
 {
  "IBM",
  "ISO",
  "ASCII",
  (char *)NULL
 };

char *long_charsets[]=
 {
  "IBM - geeignet fr Telix, Telemate...",
  "ISO - geeignet fr Windows, Amiga und Unix",
  "ASCII - wenn sonst nichts mehr hilft",
  (char *)NULL
 };

void screent::flushkeys()
 {
  while(readkey()!=0);
 }

void screent::convert_out_str(char *in, char *out)
 {
  strcpy(out, in); // out is assumed to be large enough

  switch(app_out_charset)
   {
   case IBM:
    switch(hw_out_charset)
     {
     case ISO:
      ibm2iso(out, strlen(out));
      break;

     case ASCII:
      ibm2ascii(in, out, TRUE);
      break;

     default: 
      break;
     }
    break;

   case ISO:
    switch(hw_out_charset)
     {
     case IBM:
      iso2ibm(out, strlen(out));
      break;

     case ASCII:
      iso2ascii(in, out, TRUE);
      break;

     default:
      break;
     }
    break;

   default:
    break;
   }
 }

static char boxchars[5][6]=
 {
  { '', '', '', '', '', '' },
  { '', '', '', '', '', '' },
  { '', '', '', '', '', '' },
  { '', '', '', '', '', '' },
  { '', '', '', '', '', '' }
 };

void screent::box(byte x1, byte y1, byte x2, byte y2,
  frametypet frametype, const char *title, bool clear)
 {
  char *bchars=boxchars[frametype-1];
  byte x, y;

  for(y=y1; y<=y2; y++)
   {
    x=x1;
    setcursor(x, y);
    if(y==y1)
     {
      printf("%c", bchars[0]);
      for(x++; x<x2; x++) printf("%c", bchars[4]);
      printf("%c", bchars[1]);

      if(title!=NULL && title[0]!=0)
       {
        byte l1, l2;
        char t[1000];

        snprintf(t, 1000, " %s ", title);
        x=x1;
        l1=strlen(t);
        l2=x2-x1+1;
        if(l1>l2-2) l1=l2-2;
        x+=(l2-l1)/2;
        setcursor(x, y);
        t[l1]=0;
        printf("%s", t);
       }
     }
    else if(y==y2)
     {
      printf("%c", bchars[3]);
      for(x++; x<x2; x++) printf("%c", bchars[4]);
      printf("%c", bchars[2]);
     }
    else
     {
      printf("%c", bchars[5]);
      if(clear)
        for(x++; x<x2; x++) putch(' ');
      else
        setcursor(x2, y);
      printf("%c", bchars[5]);
     }
   }
 }

word screent::doreadkey()
 {
  word ch;

  asyncio.flush_output();

  ch=readkey();
  if(ch==0)
   {
    waitkey();
    ch=readkey();
   }

  return ch;
 }

word ansiscreent::readkey()
 {
  char ch;

  if(bposwrite!=0)
   {
    ch=buffer[bposread++];
    if(bposread==bposwrite)
      bposread=bposwrite=0;
    goto keyfound;
   }

  if(readchar(&ch))
   {
    if(ch==27)
      return processesc();
   }
  else
    return 0;

 keyfound:
  switch(ch)
   {
   case 0:
    return Key_C_2;

   case 13:
    return Key_Ent;

   case 127:
    return Key_BS;

   case 27:
    return Key_Esc;

   default:
    if(lo(ch)!=0)
     {
      switch(hw_in_charset)
       {
       case ISO:
        switch(app_in_charset)
         {
         case IBM:
          iso2ibm((char *)&ch, 1);
          break;

         default:
          break;
         }
        break;

       case IBM:
        switch(app_in_charset)
         {
         case ISO:
          ibm2iso((char *)&ch, 1);
          break;

         default:
          break;
         }
        break;

       default:
        break;
       }
     }
   }
  return ch;
 }

void ansiscreent::setattr(byte type, bool marked)
 {
  printf("\033[0m");

  if(usecolor)
    switch(type)
     {
     case 1:
      setcolor(col1v+(marked?8:0), col1h);
      break;

     case 2:
      setcolor(col2v+(marked?8:0), col2h);
      break;

     case 3:
      setcolor(col3v+(marked?8:0), col3h);
      break;

     case 10:
      setcolor(marked?4:0, 7);
      //printf("\033[7m");
      //if(marked) printf("\033[4m");
      break;

     case 11:
      setcolor(marked?4:0, 6);
      break;

     case 0xff:
      if(marked) printf("\033[1m");
      break;

     default:
      logf.internalerror("ansiscreent::setattr", "unbekanntes Attribut");
     }
  else
    switch(type)
     {
     case 1:
     case 2:
     case 3:
     case 0xff:
      if(marked) printf("\033[1m");
      break;

     case 10:
      printf("\033[7m");
      if(marked) printf("\033[4m");
      break;

     case 11:
      printf("\033[7m");
      printf("\033[4m");
      break;

     default:
      logf.internalerror("ansiscreent::setattr", "unbekanntes Attribut");
     }
 }

void ansiscreent::setcolor(byte v, byte h)
 {
  bool hi=((v&8)!=0);
  printf("\033[m");
  if(v<16) printf("\033[3%um", v&7);
  if(h<16) printf("\033[4%um", h&7);
  if(hi) printf("\033[1m");
 }

void ansiscreent::movecursor(signed char x, signed char y)
 {
  if(x<0)
    printf("\033[%uD", -x);
  else if(x>0)
    printf("\033[%uC", x);

  if(y<0)
    printf("\033[%uA", -y);
  else if(y>0)
    printf("\033[%uB", y);
 }

void ansiscreent::getsize()
 {
  int i;
  char answer[20], ch, *tptr;

  setcursor(1, 100);
  printf("\033[6n");
  sleep(1);

  do { ch=readkey(); } while(ch!=0 && ch!=0x1b);

  if(ch==0)
   {
    printf("\r                   \r");
    return;
   }

  i=0;
 schleife:
  answer[i]=readkey();
  if(i!=18 && answer[i]!=0 && answer[i]!='R')
   {
    i++;
    goto schleife;
   }
  i++;
  answer[i]=0;

  if(answer[0]!='[') return;

  tptr=strchr(answer, ';');
  if(tptr==NULL) return;

  *(tptr++)=0;

  sizey=atoi(answer+1);
 }

bool ansiscreent::open()
 {
  sizex=80;
  sizey=24;
  hw_out_charset=ASCII;
  return FALSE;
 }

bool ansiscreent::close()
 {
  return FALSE;
 }

struct keymapt
 {
  char sequence[10];
  word key;
 } keymaps[]=
 {
   { "[A", Key_Up },
   { "[B", Key_Dwn },
   { "[C", Key_Rgt },
   { "[D", Key_Lft },
   { "[H", Key_Home },
   { "[W", Key_Home },
   { "[K", Key_End },
   { "[F", Key_End },
   { "[Q", Key_End },
   { "[Y", Key_PgUp },
   { "[S", Key_PgDn },
   { "[1~", Key_Home },
   { "[2", Key_Home },
   //{ "[3", Key_PgUp }, // conflicts with Del
   { "[4~", Key_End },
   { "[5~", Key_PgUp },
   { "[6~", Key_PgDn },
   { "[5z", Key_PgUp },
   { "[6z", Key_PgDn },
   { "[7~", Key_Home },
   { "[8~", Key_End },
   { "[[A", Key_F1 },
   { "[[B", Key_F2 },
   { "[[C", Key_F3 },
   { "[[D", Key_F4 },
   { "[[E", Key_F5 },
   { "[11~", Key_F1 },
   { "[12~", Key_F2 },
   { "[13~", Key_F3 },
   { "[14~", Key_F4 },
   { "[15~", Key_F5 },
   { "[17~", Key_F6 },
   { "[18~", Key_F7 },
   { "[19~", Key_F8 },
   { "[20~", Key_F9 },
   { "[21~", Key_F10 },
   { "[23~", Key_F11 },
   { "[24~", Key_F12 },
   { "[3~", Key_Del }, // Entf
   { "[x", Key_Up },
   { "[r", Key_Dwn },
   { "[v", Key_Rgt },
   { "[t", Key_Lft },
   { "[w", Key_Home },
   { "[q", Key_End },
   { "[e", Key_End },
   { "[y", Key_PgUp },
   { "[s", Key_PgDn },
   { "[M", Key_Ent },
   { "[h", Key_Del }, // Entf
   { "", 0 }
 };

word ansiscreent::processesc()
 {
  int i;

  for(i=0; i<10; i++) buffer[i]=0;

  while(TRUE)
   {
    char ch;

    if(!readchar(&ch))
     {
      processescwait();
      if(!readchar(&ch))
        goto abort;
     }

    if(ch==0) ch='*';
    buffer[bposwrite++]=ch;

    switch(bposwrite)
     {
     case 1:
      switch(ch)
       {
       case '[':
        break;

       case 'O':
        buffer[0]='[';
        break;

       default:
        goto abort;
       }
      break;

     case 2:
     case 3:
     case 4:
     case 5:
      for(i=0; keymaps[i].key!=0; i++)
        if(strcmp(buffer, keymaps[i].sequence)==0)
         {
          bposread=bposwrite=0;
          return keymaps[i].key;
         }
      break;

     default:
      goto abort;
     }
   }

 abort:
  return Key_Esc;
 }

#if 0
int remoteconsolet::printf(const char *format, ...)
 {
  va_list args;
  char buffer[2000];
  int chars, i;
  byte message[3000];

  va_start(args, format);

  chars=vsnprintf(buffer, 2000, format, args);

  va_end(args);

  if(cursorx+chars>(sizex+1)) return 0;

  for(i=0; i<chars; i++)
   {
    data[cursory*sizex+cursorx+i].ch=buffer[i];
    data[cursory*sizex+cursorx+i].attr=attr;
    message[i+1]=buffer[i];
   }

  cursorx+=chars;

  message[0]='r';
  sendmsg(message, chars+1); // printf

  return chars;
 }

void remoteconsolet::putch(char ch)
 {
  byte message[100];

  if(cursorx>sizex) return;

  data[cursory*sizex+cursorx].ch=ch;
  data[cursory*sizex+cursorx].attr=attr;

  cursorx++;

  message[0]='p';
  message[1]=ch;
  sendmsg(message, 2); // putch
 }

word remoteconsolet::readkey()
 {
  return 0;
 }

bool remoteconsolet::readchar(char *ch)
 {

  return FALSE;
 }

void remoteconsolet::waitkey()
 {
 }

void remoteconsolet::clrscr()
 {
  word x, y; 

  for(y=0; y<sizey; y++)
    for(x=0; x<sizex; x++)
     {
      data[y*sizex+x].ch=' ';
      data[y*sizex+x].attr=attr;
     }

  cursorx=cursory=1;

  sendmsg((byte *)"c", 1); // clrscr
 }

void remoteconsolet::setattr(byte type, bool marked)
 {
  byte message[100];

  attr=0; // TODO

  message[0]='a';  // set attr
  message[1]=attr;

  sendmsg(message, 2);
 }

void remoteconsolet::setcolor(byte v, byte h)
 {
  byte message[100];

  attr=0; // TODO

  message[0]='a';  // set attr
  message[1]=attr;

  sendmsg(message, 2);
 }

void remoteconsolet::setcursor(byte x, byte y)
 {
  byte message[100];

  cursorx=x;
  cursory=y;

  message[0]='s';  // set cursor
  message[1]=cursorx;
  message[2]=cursory;

  sendmsg(message, 3);
 }

void remoteconsolet::setcursorx(byte x)
 {
  byte message[100];

  cursorx=x;

  message[0]='s';  // set cursor
  message[1]=cursorx;
  message[2]=cursory;

  sendmsg(message, 3);
 }

void remoteconsolet::movecursor(signed char x, signed char y)
 {
  byte message[100];

  cursorx+=x;
  cursory+=y;

  if(cursorx<1) cursorx=1;
  if(cursory<1) cursory=1;
  if(cursorx>sizex) cursorx=sizex;
  if(cursory>sizey) cursory=sizey;

  message[0]='s';  // set cursor
  message[1]=cursorx;
  message[2]=cursory;

  sendmsg(message, 3);
 }

void remoteconsolet::deleteline()
 {
  sendmsg((byte *)"e", 1); // delete line
 }

void remoteconsolet::insertline()
 {
  sendmsg((byte *)"i", 1); // insert line
 }

void remoteconsolet::cleareol()
 {

 }

bool remoteconsolet::open()
 {
  int i;

  screent::open();

  data=(cellt *)malloc(sizex*sizey);
  if(data==NULL)
    logf.internalerror("remoteconsolet::open", "malloc failed");

  for(i=0; i<MAXPORTS; i++)
    send_updates[i]=0;

  clrscr();

  return FALSE;
 }

void remoteconsolet::start_update_screen(word port, pid_t pid)
 {
  byte message[100];
  word x, y, i;

  send_updates[port]=pid;
  send_only=port;

  message[0]='d';  // screen size
  message[1]=sizex;
  message[2]=sizey;

  sendmsg(message, 3);

  sendmsg((byte *)"c", 1); // clrscr

  for(y=0; y<sizey; y++)
   {
    message[0]='l'; // line
    message[1]=y;
    i=2;
    for(x=0; x<sizex; x++) 
     {
      message[i++]=data[y*sizex+x].ch;
      message[i++]=data[y*sizex+x].attr;
     }
    sendmsg(message, i);
   }

  message[0]='a';  // set attr
  message[1]=attr;

  sendmsg(message, 2);

  message[0]='s';  // set cursor
  message[1]=cursorx;
  message[2]=cursory;

  sendmsg(message, 3);

  send_only=0;  
 }

void remoteconsolet::processmsg(byte *message, dword len, pid_t sourcepid)
 {

 }

void remoteconsolet::sendmsg(byte *message, dword len)
 {
  ipcmsgt *msg;
  int i;

  if(send_only==0)
   {
    for(i=0; i<MAXPORTS; i++)
      if(send_updates[i]!=0) break;

    if(i==MAXPORTS) return;
   }

  msg=(ipcmsgt *)malloc(sizeof(ipcmsgt)+len);
  if(msg==NULL) 
    logf.internalerror("remoteconsolet::sendmsg", "malloc failed");

  memcpy(((byte *)msg)+sizeof(ipcmsgt), message, len);

  msg->typ=IPCMSG_SCREEN;
  msg->len=len;

  if(send_only==0)
   {
    for(i=0; i<MAXPORTS; i++)
      if(send_updates[i]!=0)
        if(kill(send_updates[i], 0)!=0 || ipc.send(i, send_updates[i], msg))
          send_updates[i]=0;
   }
  else
   {
    ipc.send(send_only, send_updates[send_only], msg);
   }

  free(msg);
 }

bool remoteconsolet::close()
 {
  if(data!=NULL)
   {
    free(data);
    data=(cellt *)NULL;
   }

  return FALSE;
 }

void remoteconsolet::set_screensize(int new_sizex, int new_sizey)
 {

 }
#endif

#define ALARMTIME 20

void devterminalt::flush()
 {
  if(sigalarm) return;
  alarm(ALARMTIME); // wg. Haenger!
  tcflush(fd, TCOFLUSH);
  alarm(0);
 }

int devterminalt::printf(const char *format, ...)
 {
  va_list args;
  char buffer[2000];
  int chars;

  va_start(args, format);
  chars=vsnprintf(buffer, 2000, format, args);
  va_end(args);

  if(hw_out_charset==app_out_charset)
    asyncio.write(fd, buffer, chars);
  else
   {
    char buffer2[4000];
    convert_out_str(buffer, buffer2);
    chars=strlen(buffer2);
    asyncio.write(fd, buffer2, chars);
   }

  if(sigalarm) return chars;

  return chars;
 }

void devterminalt::putch(char ch)
 {
  asyncio.write(fd, &ch, 1);
 }

void devterminalt::processescwait()
 {
  vector<int> handles;
  handles.push_back(fd);
  asyncio.input(handles);
 }

bool devterminalt::readchar(char *ch)
 {
  if(read(fd, ch, 1)!=1)
    return FALSE;

  return TRUE;
 }

void devterminalt::waitkey()
 {
  vector<int> handles;
  handles.push_back(fd);
  asyncio.input(handles);
 }

bool devterminalt::assign_to_file(const char *filename)
 {
  bufsize=100;
  buf=0;

  if(filename[0]==0)
   {
    fd=1;
   }
  else
   {
    fd=::open(filename, O_RDWR);
    if(fd<0)
      logf.printf("devterminalt::assign_to_file", "Failed to open \"%s\"", filename);
   }

  sigalarm=FALSE;
  return fd<0;
 }

bool devterminalt::open()
 {
  struct termios t;

  asyncio.make_async(fd);

  ioctl(fd, TCGETA, &told);

  tcgetattr(fd, &t);

  t.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL|IXON);
  t.c_oflag &= ~OPOST;
  t.c_lflag &= ~(ECHO|ECHONL|ICANON|ISIG|IEXTEN);
  t.c_cflag &= ~(CSIZE|PARENB);
  t.c_cflag |= CS8;

  tcsetattr(fd, 0, &t);

  return ansiscreent::open();
 }

bool devterminalt::close()
 {
  asyncio.flush_output();
  ansiscreent::close();
  ioctl(fd, TCSETA, &told);
  if(fd!=1) ::close(fd);
  return FALSE;
 }

int consolet::printf(const char *format, ...)
 {
  va_list args;
  char buffer[2000];
  int chars;

  va_start(args, format);
  chars=vsnprintf(buffer, 2000, format, args);
  va_end(args);

  if(hw_out_charset==app_out_charset)
    asyncio.write(fd, buffer, chars);
  else
   {
    char buffer2[4000];
    convert_out_str(buffer, buffer2);
    chars=strlen(buffer2);
    asyncio.write(fd, buffer2, chars);
   }

  return chars;
 }

void consolet::set_app_out_charset(charsett charset)
 {
  app_out_charset=charset;

  if(linuxconsole)
   {
    if(charset==IBM && hw_out_charset!=IBM)
     {
      printf("%c(U", 27);
      /*printf("%c[?25l", 27);*/ /* Cursor aus? */
      hw_out_charset=IBM;
     }
    else if(charset==ISO && hw_out_charset!=ISO)
     {
      printf("%c(B", 27);
      /* printf("%c[?25h", 27); */
      hw_out_charset=ISO;
     }
   }
 }

bool consolet::open()
 {
  int ioctl_fd;

  if(devterminalt::open()) return TRUE;

  #ifdef OS_WIN32
  hw_out_charset=hw_in_charset=IBM;
  #else
  hw_out_charset=hw_in_charset=ISO;

  if(linuxconsole) set_app_out_charset(IBM);
  #endif

  sizey=25;
  sizex=80;

  ioctl_fd=fd;
  if(ioctl_fd==1) ioctl_fd=0;

  #ifdef TIOCGSIZE
  struct ttysize getit;

  if(ioctl(ioctl_fd, TIOCGSIZE, &getit)!= -1)
   {
    sizex=getit.ts_cols;
    sizey=getit.ts_lines;
   }
  #else
  #ifdef TIOCGWINSZ
  struct winsize getit;

  if(ioctl(ioctl_fd, TIOCGWINSZ, &getit)!= -1)
   {
    sizex=getit.ws_col;
    sizey=getit.ws_row;
   }
  #endif
  #endif

  if(sizex<80 || sizex>200) sizex=80;

  if(sizey<24 || sizey>200) sizey=24;

  return FALSE;
 }

bool consolet::close()
 {
  set_app_out_charset(ISO);
  if(!sigalarm) tcdrain(fd);
  return devterminalt::close();
 }

bool consolet::assign(const char *device)
 {
  char filename[DIRLEN];

  if(strnicmp(device, "/dev/", 5)==0) device+=5;
  
  if(device[0]==0 || stricmp(device, "tty")==0)
   {
    filename[0]=0;
    #ifdef OS_WIN32
    linuxconsole=FALSE;
    #else
    linuxconsole=(stricmp(getenv("TERM"), "linux")==0);
    #endif
   }
  else if(isdigit(device[0]))
   {
    snprintf(filename, DIRLEN, "/dev/tty%d", atoi(device));
    linuxconsole=TRUE;
   }
  else if(device[0]=='/')
   {
    linuxconsole=FALSE;
    strmaxcpy(filename, device, DIRLEN-1);
   }
  else
   {
    linuxconsole=FALSE;
    snprintf(filename, DIRLEN, "/dev/%s", device);
   }

  return devterminalt::assign_to_file(filename);
 }

/* Different names for the same beast. */
#ifndef TIOCMODG			/* BSD 4.3 */
#  ifdef TIOCMGET
#    define TIOCMODG TIOCMGET		/* Posix */
#  else
#    ifdef MCGETA
#      define TIOCMODG MCGETA		/* HP/UX */
#    endif
#  endif
#endif

#ifndef TIOCMODS
#  ifdef TIOCMSET
#    define TIOCMODS TIOCMSET
#  else
#    ifdef MCSETA
#      define TIOCMODS MCSETA
#    endif
#  endif
#endif

#ifndef TIOCM_CAR			/* BSD + Posix */
#  ifdef MDCD
#    define TIOCM_CAR MDCD		/* HP/UX */
#  endif
#endif

/* Set RTS line. Sometimes dropped. Linux specific? */

void m_setrts(int fd)
 {
  #ifdef OS_UNIX
  int mcs;

  ioctl(fd, TIOCMODG, &mcs);
  mcs|=TIOCM_RTS;
  ioctl(fd, TIOCMODS, &mcs);
  #endif
 }

/*
 * Set the line status so that it will not kill our process
 * if the line hangs up.
 */

void m_nohang(int fd)
 {
  struct termios sgg;

  tcgetattr(fd, &sgg);
  sgg.c_cflag |= CLOCAL;
  tcsetattr(fd, TCSANOW, &sgg);
 }

/*
 * Flush the buffers
 */

#if 0
void m_flush(int fd)
 {
  ioctl(fd, TCFLSH, (void *)2);
 }
#endif

void comportt::dtrtoggle()
 {
  if(usesocket)
   {

   }
  else
   {
    bool carrier=dcd();
    struct termios tty, old;
    alarm(ALARMTIME); // wg. Haenger!
    tcgetattr(fd, &tty);
    tcgetattr(fd, &old);
    cfsetospeed(&tty, B0);
    cfsetispeed(&tty, B0);
    tcsetattr(fd, TCSANOW, &tty);
    tcsetattr(fd, TCSANOW, &old);
    tcsetattr(fd, TCSANOW, &tty);
    sleep(carrier?3:1);
    if(dcd()) sleep(1);
    tcsetattr(fd, TCSANOW, &old);
    if(carrier) sleep(3);
    alarm(0);
   }
 }

bool comportt::dcd()
 {
  if(usesocket)
   { // DICKER DK! HACK in /usr/src/linux/net/inet/sock.c
    #if 0
    int i;
    socklen_t len=4;
    if(getsockopt(fd, SOL_SOCKET, 61175, &i, &len)!=0)
      logf.internalerror("comportt::dcd", "getsockopt failed");

    if(i==TCP_ESTABLISHED || i==TCP_SYN_SENT || i==TCP_SYN_RECV)
      return TRUE;
    #endif

    return FALSE;
   }
  else
   {
    #ifdef OS_WIN32
    return FALSE;
    #else
    int mcs;
    bool error;

    alarm(ALARMTIME); // wg. Haenger!
    error=(ioctl(fd, TIOCMODG, &mcs)!=0);
    alarm(0);
    return (mcs&TIOCM_CAR?TRUE:FALSE) && !error;
    #endif
   }
 }

void comportt::setparms(dword baudr, char par, char bits)
 {
  if(!usesocket)
   {
    int spd=-1;
    struct termios tty;

    tcgetattr(fd, &tty);

    switch(baudr)
     {
     case 0:
      #ifdef B0
      spd=B0;
      #else
      spd=0;
      #endif
      break;

     case 50:
      spd=B50;
      break;

     case 75:
      spd=B75;
      break;

     case 300:
      spd=B300;
      break;

     case 600:
      spd=B600;
      break;

     case 1200:
      spd=B1200;
      break;

     case 2400:
      spd=B2400;
      break;

     case 4800:
      spd=B4800;
      break;

     case 9600:
      spd=B9600;
      break;

     case 19200:
      #ifdef B19200
      spd=B19200;
      #else
      #ifdef EXTA
      spd=EXTA;
      #else
      spd=B9600;
      #endif
      #endif
      break;

     case 38400:
      #ifdef B38400
      spd=B38400;
      #else
      #ifdef EXTA
      spd=EXTA;
      #else
      spd=B9600;
      #endif
      #endif	
      break;

     case 57600:
      #ifdef B57600
      spd=B57600;
      #endif;
      break;

     case 115200:
      #ifdef B115200
      spd=B115200;
      #endif
      break;
     }
  
    if(spd!=-1 && baudr!=0)
     {
      cfsetospeed(&tty, (speed_t)spd);
      cfsetispeed(&tty, (speed_t)spd);
     }

    switch(bits)
     {
     case 5:
     case '5':
      tty.c_cflag=(tty.c_cflag&~CSIZE)|CS5;
      break;

     case 6:
     case '6':
      tty.c_cflag=(tty.c_cflag&~CSIZE)|CS6;
      break;

     case 7:
     case '7':
      tty.c_cflag=(tty.c_cflag&~CSIZE)|CS7;
      break;
   
     case 8:
     case '8':
     default:
      tty.c_cflag=(tty.c_cflag&~CSIZE)|CS8;
      break;
     }
   
    /* Set into raw, no echo mode */
  
    tty.c_iflag&=~(IGNBRK|IGNCR|INLCR|ICRNL|IUCLC|IXANY|IXON|IXOFF|INPCK|ISTRIP);
    tty.c_iflag|=(BRKINT|IGNPAR);
    tty.c_oflag&=~OPOST;
    tty.c_lflag=~(ICANON|ISIG|ECHO|ECHONL|ECHOE|ECHOK);
    tty.c_cflag|=CREAD|CRTSCTS;
    tty.c_cc[VMIN]=1;
    tty.c_cc[VTIME]=5;

    tty.c_cflag&=~(PARENB|PARODD);
    if(par=='E')
      tty.c_cflag|=PARENB;
    else if(par=='O')
      tty.c_cflag|=PARODD;

    tcsetattr(fd, TCSANOW, &tty);

    m_setrts(fd);
   }
 }

bool comportt::assign(const char *device)
 {
  char filename[DIRLEN];

  if(device[0]==0)
    return TRUE;
  else if(isdigit(device[0]))
   {
    #ifdef OS_WIN32
    snprintf(filename, DIRLEN, "/dev/com%d", atoi(device));
    #else
    snprintf(filename, DIRLEN, "/dev/cua%d", atoi(device));
    #endif
   }
  else if(device[0]=='/')
    strmaxcpy(filename, device, DIRLEN-1);
  else
    snprintf(filename, DIRLEN, "/dev/%s", device);

  usesocket=FALSE;
  return devterminalt::assign_to_file(filename);
 }

bool comportt::opensocket()
 {
  struct sockaddr_in sin;
 
  fd=socket(AF_INET, SOCK_STREAM, 0);
  if(fd<0)
   {
    logf.printf("comportt::opensocket", "socket failed!");
    return TRUE;
   }
  
  sin.sin_family=AF_INET;
  sin.sin_port=mword(hi(port), lo(port));
  sin.sin_addr.s_addr=addr; 

  if(connect(fd, (struct sockaddr *)&sin, sizeof (sin))<0)
   {
    logf.printf("comportt::opensocket", "connect failed!");
    return TRUE;
   }
  return FALSE;
 }

