#ifndef DBOX_OLD_TERM_H

#define DBOX_OLD_TERM_H

#include <termio.h>
#include <charsett.h>
#include <interlib.h>

#include <libdbox.h>
#include <asyncio.h>

extern char *charsets[];
extern char *long_charsets[];

enum frametypet { BOXFRAMENON, BOXFRAMESNG, BOXFRAMEDBL, BOXFRAMEVER,
                  BOXFRAMEHOR, BOXFRAMEHVY };

class screent
 {
 public:
  void convert_out_str(char *in, char *out);

  virtual int printf(const char *format, ...)=0;
    /* mit Zeichensatzkonvertierung */

  virtual void putch(char ch)=0;
    /* ohne Zeichensatzkonvertierung */

  virtual word readkey()=0;
    /* Mit Konvertierung */

  virtual word doreadkey();
    /* Mit Konvertierung und Warten
       0xffff=Bitte beenden! */

  virtual void flushkeys();

  virtual bool readchar(char *ch)=0;
  virtual void waitkey()=0;
  virtual void clrscr()=0;
  virtual void setattr(byte type, bool marked)=0;
    // 1=Normaler Text1
    // 2=...
    //
    // 10=Buttontext
    // 11=gewhlter Buttontext
    //
    //

  virtual void setcolor(byte v, byte h)=0;
  virtual void setcursor(byte x, byte y)=0; /* 1:1 = links oben */
  virtual void setcursorx(byte x)=0;
  virtual void movecursor(signed char x, signed char y)=0;
  virtual void deleteline()=0;
  virtual void insertline()=0;
  virtual void cleareol()=0;
  virtual bool open() { return FALSE; }
  virtual bool close() { return FALSE; }

  virtual void box(byte x1, byte y1, byte x2, byte y2,
    frametypet frametype, const char *title, bool clear);

  byte sizex, sizey;

  charsett app_in_charset, app_out_charset,
           hw_in_charset, hw_out_charset;
    
  virtual void set_app_out_charset(charsett charset) 
   { app_out_charset=charset; }
   
  virtual void set_hw_charset(charsett charset)
   { hw_in_charset=hw_out_charset=charset; }
   
  virtual void set_usecolor(bool new_usecolor)
   { usecolor=new_usecolor; }
   
  virtual void set_screensize(int new_sizex, int new_sizey)
   { sizex=new_sizex; sizey=new_sizey; }

  bool usecolor;
  byte col1v, col2v, col3v, col1h, col2h, col3h;

  screent()
   {
    sizex=80;
    sizey=24;
    app_in_charset=app_out_charset=hw_in_charset=hw_out_charset=IBM;
    usecolor=TRUE;
   }
 };

class ansiscreent:public screent
 {
 public:
  virtual word readkey();

  virtual void clrscr()
   {
    printf("\033[2J");
   }

  virtual void setattr(byte type, bool marked);

  virtual void setcolor(byte v, byte h);

  virtual void setcursor(byte x, byte y)
   {
    printf("\033[%u;%uH", y, x);
   }

  virtual void setcursorx(byte x)
   {
    printf("\033[;%uH", x);
   }

  virtual void movecursor(signed char x, signed char y);

  virtual void deleteline()
   {
    printf("\033[1M");
   }

  virtual void insertline()
   {
    printf("\033[1L");
   }

  virtual void cleareol()
   {
    printf("\033[K");
   }

  virtual void getsize();

  virtual bool open();
  virtual bool close();

  ansiscreent() { bposread=bposwrite=0; }

  virtual void processescwait() { }

 protected:
  char buffer[10];
  byte bposread, bposwrite;

  word processesc();
 };

#if 0
class remoteconsolet:public screent
 {
 public:
  virtual int printf(const char *format, ...);
  virtual void putch(char ch);
  virtual word readkey();
  virtual bool readchar(char *ch);
  virtual void waitkey();
  virtual void clrscr();
  virtual void setattr(byte type, bool marked);
  virtual void setcolor(byte v, byte h);
  virtual void setcursor(byte x, byte y);
  virtual void setcursorx(byte x);
  virtual void movecursor(signed char x, signed char y);
  virtual void deleteline();
  virtual void insertline();
  virtual void cleareol();
  virtual bool open();
  virtual bool close();
  virtual void set_screensize(int new_sizex, int new_sizey);
  virtual void processmsg(byte *message, dword len, pid_t sourcepid);
  virtual void start_update_screen(word port, pid_t pid);
  virtual void sendmsg(byte *message, dword len);
  
  struct cellt
   {
    byte ch, attr;
   } *data;
   
  byte attr, cursorx, cursory;
  
  pid_t send_updates[MAXPORTS];
  word send_only;
   
  remoteconsolet()
   {
    data=(cellt *)NULL;
    attr=0xff;
    send_only=0;
   }
 };
#endif

class devterminalt:public ansiscreent
 {
 public:
  virtual void flush();
  virtual int printf(const char *format, ...);
  virtual void putch(char ch);
  virtual bool readchar(char *ch);
  virtual void waitkey();
  virtual bool assign_to_file(const char *filename);
  virtual bool open();
  virtual bool close();
  virtual void processescwait();

  int fd;
  struct termio told, tnew;
  bool sigalarm;
  word bufsize, buf;
 };

class consolet:public devterminalt
 {
 public:
  bool linuxconsole;
  
  virtual int printf(const char *format, ...);

  virtual bool open();
  virtual bool close();
  virtual bool assign(const char *device);

  virtual void set_app_out_charset(charsett charset);
  virtual void set_hw_charset(charsett charset)
   { if(!linuxconsole) hw_in_charset=hw_out_charset=charset; }
 };

class comportt:public devterminalt
 {
 public:
  internetaddresst addr;
  word port;
  bool usesocket;

  virtual void dtrtoggle();
  virtual bool dcd();
  virtual void setparms(dword baudr, char par, char bits);
  virtual bool assign(const char *device);
  virtual bool opensocket();
 };

#endif // DBOX_OLTERM_H

