#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <oldterm.h>
#include <keys.h>
#include <ui.h>
#include <nls.h>

menut::menut(screent *_screen)
 {
  draw_all=TRUE;
  last=selected=mes=(met *)NULL;
  screen=_screen;
  x=y=0;
 }

menut::~menut()
 {
  met *me, *next;
  me=mes;
  while(me!=NULL)
   {
    next=me->next;
    if(me->free_me) delete me;
    me=next;
   }
 }

void menut::result(int id)
 {
  result_id=id;
  result_available=TRUE;
 }

void menut::draw()
 {
  met *me;

  for(me=mes; me!=NULL; me=me->next)
    me->draw();
 }

int menut::menu()
 {
  word key;
  met *me;

  result_available=FALSE;

  if(selected==NULL)
   {
    selected=mes;
    while(selected!=NULL && !selected->selectable)
      selected=selected->next;
   }

  if(draw_all)
   {
    draw();
    draw_all=FALSE;
   }

  if(selected==NULL)
    screen->setcursor(1, 1);
  else
    selected->select();

  while(TRUE)
   {
    key=screen->doreadkey();

    if(key==0) continue;

    me=mes;
    while(me!=NULL && me->process_key(key))
      me=me->next;

    if(result_available)
      return result_id;

    if(me!=NULL) continue;

    switch(key)
     {
     case 0xffff:
      return MENU_QUIT;

     case Key_Esc:
      return MENU_ABORT;

     case Key_F1:
      return MENU_HELP;

     case Key_Up:
     case Key_Lft:
      do_up();
      break;
 
     case Key_Tab:
     case Key_Dwn:
     case Key_Rgt:
     case Key_Spa:
      do_down();
      break;

     case Key_Ent:
     case Key_End:
     case Key_Home:
     case Key_BS:
      break;

     default:
      break;
     }
   }
 }

void menut::do_down()
 {
  met *newselected;

  if(selected!=NULL)
   {
    newselected=selected->next;
    if(newselected==NULL) newselected=mes;

    while(newselected!=NULL && !newselected->selectable)
      newselected=newselected->next;

    select(newselected);
   }
 }

void menut::do_up()
 {
  met *newselected;

  if(selected!=NULL)
   {
    newselected=selected->prev;
    if(newselected==NULL) newselected=last;

    while(newselected!=NULL && !newselected->selectable)
      newselected=newselected->prev;

    select(newselected);
   }
 }

void menut::add(met *me)
 {
  me->menu=this;

  if(last==NULL)
    mes=last=me;
  else
   {
    me->prev=last;
    last->next=me;
    last=me;
   }
 }
 
void menut::select(met *me)
 {
  met *oldselected;

  if(selected==me) return;

  // 1st step: Deselect previously selected item

  if(selected!=NULL)
   {
    oldselected=selected;
    selected=(met *)NULL;
    oldselected->select();
   }

  if(me!=NULL && me->selectable)
   {
    selected=me;
    selected->select();
   }
 }

met::met()
 {
  selectkey=0;
  x=y=sizex=sizey=1;
  selectable=FALSE;
  disabled=FALSE;
  ignore=FALSE;
  free_me=FALSE;
  next=prev=(met *)NULL;
  id=0;
 }
   
void met::draw()
 {
 }
  
met::~met() 
 {
 }

void met::select()
 {
  draw();
 }

bool met::process_key(word key)
 {
  if(key==selectkey)
   {
    menu->select(this);
    return FALSE;
   }
    
  return TRUE;
 }

editmet::editmet():met()
 {
  maxlen=0;
  selectable=TRUE;
  posx=scrollx=0;
 }

void editmet::select()
 {
  bool selected=(menu->selected==this);

  if(selected)
   {
    posx=str_len();
    if(posx>sizex-1)
      scrollx=posx-sizex+1;
    else
      scrollx=0;
   }
  else
    posx=scrollx=0;

  met::select();
 }
   
void editmet::draw()
 {
  bool selected=(menu->selected==this);
  const char *tptr;
  int i;

  menu->screen->setcursor(x+menu->x, y+menu->y);
  menu->screen->setattr(disabled?1:(selected?11:10), FALSE);
  menu->screen->putch(selected?'>':' ');

  tptr=str_text()+scrollx;
  i=sizex;
  while(*tptr!=0 && i!=0)
   {
    if(*tptr>=' ' || *tptr<0)
      menu->screen->printf("%c", *tptr);
    else
     {
      menu->screen->setattr(selected?11:10, TRUE);
      menu->screen->printf("*");
      menu->screen->setattr(selected?11:10, FALSE);
     }
    i--;
    tptr++;
   }

  for(; i>0; i--)
    menu->screen->putch('_');

  menu->screen->putch(selected?'<':' ');
  menu->screen->setattr(1, FALSE);

  if(selected) setcursor();
 }

void editmet::setcursor()
 {
  menu->screen->setcursor(x+menu->x+posx-scrollx+1, y+menu->y);
 }

bool editmet::checkchar(char ch)
 {
  return FALSE;
 }

bool editmet::process_key(word key)
 {
  bool selected=(menu->selected==this);

  if(!selected)
    return met::process_key(key);

  switch(key)
   {
   case Key_Lft:
    if(posx!=0)
     {
      posx--;
      if(posx<scrollx)
       {
        scrollx--;
        draw();
       }
      else
        setcursor();
     }
    break;

   case Key_Rgt:
    if(posx!=str_len())
     {
      posx++;
      if(posx-scrollx>=sizex)
       {
        scrollx++;
        draw();
       }
      else
        setcursor();
     }
    break;

   case Key_Ent:
    menu->do_down();
    break;

   case Key_End:
   case Key_C_E:
    if(posx<str_len())
     {
      posx=str_len();
      if(posx-scrollx>=sizex-1)
       {
        scrollx=posx-sizex+1;
        draw();
       }
      else
        setcursor();
     }
    break;

   case Key_Home:
   case Key_C_A:
    if(posx!=0)
     {
      posx=0;
      if(scrollx!=0)
       {
        scrollx=0;
        draw();
       }
      else
        setcursor();
     }
    break;

   case Key_BS:
    if(posx!=0)
     {
      str_remove();
      posx--;

      if(posx<scrollx)
       {
        scrollx--;
        draw();
       }
      else
       {
        setcursor();
        menu->screen->setattr(11, FALSE);
        menu->screen->printf("%.*s",
          sizex-posx+scrollx, str_text()+posx);

        if(str_len()<(sizex+scrollx))
          menu->screen->printf("_");

        menu->screen->setattr(1, FALSE);
        setcursor();
       }
      break;

     default:
      char ch=lo(key);
      if((ch>=' ' && ch<127) || ch<0)
       {
        if(!checkchar(ch) && str_len()<maxlen)
         {
          str_store(ch);
          posx++;
          if(posx-scrollx>=sizex)
           {
            scrollx++;
            draw();
           }
          else
           {
            menu->screen->setattr(11, FALSE);
            menu->screen->printf("%.*s", 
              sizex-posx+scrollx+1, str_text()+posx-1);
            menu->screen->setattr(1, FALSE);
            setcursor();
           }
         }
       }
      else
        return met::process_key(key);
      break;
     }
   }

  return FALSE;
 }

editstrmet::editstrmet()
 {
  text=(char *)NULL;
 }

void editstrmet::str_store(char ch)
 {
  string tmp=text;
  tmp.insert(posx, 1, ch);
  strcpy(text, tmp.c_str());
 }

void editstrmet::str_remove()
 {
  string tmp=text;
  tmp.erase(posx-1, 1);
  strcpy(text, tmp.c_str());
 }

editstringmet::editstringmet()
 {
  text=(string *)NULL;
 }

void editstringmet::str_store(char ch)
 {
  text->insert(posx, 1, ch);
 }

void editstringmet::str_remove()
 {
  text->erase(posx-1, 1);
 }

buttonmet::buttonmet()
 {
  selectable=TRUE;
  text=(char *)NULL;
  center=TRUE;
 }
   
void buttonmet::draw()
 {
  int i, j;
  bool selected=(menu->selected==this);

  if(center)
   {
    if(strchr(text, '^')==NULL) i=0; else i=1;
    j=((int)sizex-(int)strlen(text)+i)/2;
    if(j<0) j=0;
   }
  else
    j=0;

  menu->screen->setcursor(x+menu->x, y+menu->y);
  menu->screen->setattr(disabled?1:(selected?11:10), FALSE);
  menu->screen->putch(selected?'>':' ');
  for(i=0; i<j; i++) menu->screen->putch(' ');
  const char *tptr=text;
  i=sizex-j;
  while(*tptr!=0)
   {
    if(*tptr=='^' && *(tptr+1)!=0)
     {
      tptr++;
      menu->screen->setattr(disabled?1:(selected?11:10), TRUE);
      menu->screen->printf("%c", *(tptr++));
      menu->screen->setattr(disabled?1:(selected?11:10), FALSE);
     }
    else
      menu->screen->printf("%c", *(tptr++));
    i--;
   }

  while(i>0)
   {
    menu->screen->putch(' ');
    i--;
   }

  menu->screen->putch(selected?'<':' ');
  menu->screen->setattr(1, FALSE);

  if(selected) menu->screen->setcursor(x+menu->x, y+menu->y);
 }

bool buttonmet::process_key(word key)
 {
  bool selected=(menu->selected==this);

  if(selected)
   {
    switch(key)
     {
     case Key_Ent:
      menu->result(id);
      return FALSE;
     }
   }

  return met::process_key(key);
 }

ok_buttonmet::ok_buttonmet()
 {
  text=nls["online.ok"];
  sizex=4;
  selectkey=Key_C_Ent;
 }

abort_buttonmet::abort_buttonmet()
 {
  text=nls["online.abort"];
  sizex=7;
 }

checkboxmet::checkboxmet()
 {
  checked=FALSE;
 }

void checkboxmet::draw()
 {
  bool selected=(menu->selected==this);

  menu->screen->setcursor(x+menu->x, y+menu->y);
  menu->screen->setattr(disabled?1:(selected?11:10), FALSE);
  menu->screen->putch('[');
  menu->screen->putch(checked?'x':' ');
  menu->screen->putch(']');
  menu->screen->setattr(1, FALSE);

  if(selected) menu->screen->setcursor(x+menu->x+1, y+menu->y);
 }

bool checkboxmet::process_key(word key)
 {
  bool selected=(menu->selected==this);

  if(selected)
   {
    if(key==Key_Ent || lo(key)==' ')
     {
      checked=!checked;
      draw();
      return FALSE;
     }
   }

  return met::process_key(key);
 }

void draw_title(screent &screen, const string &title)
 {
  int y, i, j;
  char tempstr[TEXTLEN*2];

  screen.setattr(10, FALSE);
  y=title.size();
  if(y>((TEXTLEN-1)*2)) y=(TEXTLEN-1)*2;
  if(y>screen.sizex-2) y=screen.sizex-2;
  strmaxcpy(tempstr, title.c_str(), y);

  j=(screen.sizex-y)/2;
  screen.setcursor(1, 1);
  for(i=0; i<j; i++) screen.printf(" ");
  screen.printf("%s", tempstr);
  j=screen.sizex-j-y;
  for(i=0; i<j; i++) screen.printf(" ");
  screen.setattr(1, FALSE);
 }

void interactive_box(screent &screen, const string &text, bool error)
 {
  int i=max(text.size()+3, (unsigned)20);
  byte x=(screen.sizex-i)/2;
  screen.setattr(error?3:2, FALSE);
  screen.box(x, 10, x+i, 16, BOXFRAMEDBL, 
             error?nls["online.error"]:(char *)NULL, TRUE);
  screen.setcursor(x+2, 12);
  screen.printf("%s", text.c_str());

  menut menu(&screen);
  menu.x=0;
  menu.y=10;

  buttonmet okme;
  okme.x=(screen.sizex-9)/2;
  okme.y=4;
  okme.selectkey=Key_C_Ent;
  okme.sizex=10;
  okme.text=nls[error?"online.abort":"online.ok"];
  menu.add(&okme);

  do
    i=menu.menu();
  while(i==MENU_HELP);

  screen.setattr(1, FALSE);
 }

void info_box(screent &screen, const string &text)
 {
  interactive_box(screen, text, FALSE);
 }

void error_box(screent &screen, const string &error)
 {
  interactive_box(screen, 
    string(nls["online.error"])+": "+error, TRUE);
 }

