#ifndef DBOX_UI_H

#define DBOX_UI_H

#include <general.h>
#include <oldterm.h>

#define MENU_ABORT -1
#define MENU_QUIT -2
#define MENU_HELP -3

class menut
 {
 public:
  class met *mes, *selected, *last;
  screent *screen;
  unsigned x, y;
  
  menut(screent *_screen);
  virtual ~menut();
  virtual void select(class met *me);
  int menu();
  virtual void add(class met *me);
  virtual void do_up();
  virtual void do_down();
  virtual void result(int id);
  virtual void draw();
 
 protected:
  int result_id;
  bool result_available;
  bool draw_all;
 };

class met
 {
 public:
  unsigned x, y, sizex, sizey;
  bool disabled, ignore, selectable, free_me;
  word selectkey;
  int id;
  
  met();
  virtual void draw();
  virtual ~met();
  virtual void select();
  virtual bool process_key(word key);  // TRUE: key not processed
  
  friend class menut;

 protected:
  met *next, *prev;
  menut *menu;
 };

class editmet:public met
 {
 public:
  unsigned maxlen;
  
  editmet();
  virtual void draw();
  virtual bool process_key(word key);
  virtual void select();
  virtual void clear_text() 
   { str_clear(); posx=scrollx=0; }
  
 protected:
  unsigned posx, scrollx;

  virtual unsigned str_len()=0;
  virtual const char *str_text()=0;
  virtual void str_store(char ch)=0;
  virtual void str_remove()=0;
  virtual void str_clear()=0;
  virtual void setcursor();  
  virtual bool checkchar(char ch);
 };

class editstrmet:public editmet
 {
 public:
  char *text;
  
  editstrmet();
  
 protected:
  virtual unsigned str_len() { return strlen(text); }
  virtual const char *str_text() { return text; }
  virtual void str_store(char ch);
  virtual void str_clear() { text[0]=0; }
  virtual void str_remove();
 };

class editstringmet:public editmet
 {
 public:
  string *text;
  
  editstringmet();
  
 protected:
  virtual unsigned str_len() { return text->size(); }
  virtual const char *str_text() { return text->c_str(); }
  virtual void str_store(char ch);
  virtual void str_clear() { (*text)=""; }
  virtual void str_remove();
 };

class buttonmet:public met
 {
 public:
  const char *text;
  bool center;
  
  buttonmet();
  virtual void draw();
  virtual bool process_key(word key);
 };
 
class ok_buttonmet:public buttonmet
 {
 public:
  ok_buttonmet();
 };
 
class abort_buttonmet:public buttonmet
 {
 public:
  abort_buttonmet();
 };

class selectmet:public met
 {
 public:
  const char **text;
  bool center;
  unsigned posy;
  
  selectmet();
  virtual void draw();
  virtual bool process_key(word key);
  
 protected:
  unsigned scrolly;
 };
  
class checkboxmet:public met
 {
 public:
  bool checked;
  
  checkboxmet();
  virtual void draw();
  virtual bool process_key(word key);
 };
  
void draw_title(screent &screen, const string &title);
void error_box(screent &screen, const string &error);
void info_box(screent &screen, const string &text);

#endif
